static int debug = 0;
module_param(debug, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(debug, "Debug enabled or not");

static int maxqueuecount = 256;
module_param(maxqueuecount, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(maxqueuecount, "Limit queue of readbuffer (0 = no limit)");

static int major = 0;
module_param(major, int, S_IRUGO | S_IWUSR);	//0644);

static int Major = 0;
static dev_t first_device = 0;

/* table of devices that work with this driver */
static struct usb_device_id lonusb_table[] = {
	{ USB_DEVICE(GESYTEC_USB_VID, GESYTEC_LVU_PID) },
	{ USB_DEVICE(GESYTEC_USB_VID, GESYTEC_LVX2_PID) },
	{ USB_DEVICE(GESYTEC_USB_VID, GESYTEC_LVX3_PID) },
	{ USB_DEVICE(GESYTEC_USB_VID, GESYTEC_GDS_PID) },
	{ USB_DEVICE(GESYTEC_USB_VID, GESYTEC_GDS_LATTICE_PID) },
	{ USB_DEVICE(GESYTEC_USB_VID, GESYTEC_LVU2_PID) },
	{ USB_DEVICE(GESYTEC_USB_VID, GESYTEC_LVU2_LATTICE_PID) },
	{ }			/* Terminating entry */
};

MODULE_DEVICE_TABLE(usb, lonusb_table);

/* array of pointers to our devices that are currently connected */
static struct _DEVICE_EXTENSION *minor_table[MAX_DEVICES];

/* lock to protect the minor_table structure */
static DEFINE_MUTEX(minor_table_mutex);

/*
 * File operations needed when we register this driver.
 * This assumes that this driver NEEDS file operations,
 * of course, which means that the driver is expected
 * to have a node in the /dev directory. If the USB
 * device were for a network interface then the driver
 * would use "struct net_driver" instead, and a serial
 * device would use "struct tty_driver".
 */
static const struct file_operations lonusb_fops = {
	/*
	 * The owner field is part of the module-locking
	 * mechanism. The idea is that the kernel knows
	 * which module to increment the use-counter of
	 * BEFORE it calls the device's open() function.
	 * This also means that the kernel can decrement
	 * the use-counter again before calling release()
	 * or should the open() function fail.
	 *
	 * Not all device structures have an "owner" field
	 * yet. "struct file_operations" and "struct net_device"
	 * do, while "struct tty_driver" does not. If the struct
	 * has an "owner" field, then initialize it to the value
	 * THIS_MODULE and the kernel will handle all module
	 * locking for you automatically. Otherwise, you must
	 * increment the use-counter in the open() function
	 * and decrement it again in the release() function
	 * yourself.
	 */
	.owner = THIS_MODULE,

	.read = lonusb_read,
	.write = lonusb_write,
	.poll = lonusb_poll,
	.unlocked_ioctl = lonusb_unlocked_ioctl,
	.open = lonusb_open,
	.release = lonusb_release,
};

// usb class driver info in order to get a minor number from the usb core,
// and to have the device registered with devfs and the driver core

static struct usb_class_driver lonusb_class = {
	.name = DRIVER_DEV_NAME "%d",
	.fops = &lonusb_fops,
	.minor_base = LONUSB_MINOR_BASE,	// ignored if CONFIG_USB_DYNAMIC_MINORS configuration option has been enabled for the kernel
	// see minor_base at http://www.makelinux.net/ldd3/chp-13-sect-4
};

/* usb specific object needed to register this driver with the usb subsystem */
static struct usb_driver lonusb_driver = {
	.name = DRIVER_MOD_NAME,
	.probe = lonusb_probe,
	.disconnect = lonusb_disconnect,
	.id_table = lonusb_table,
	.no_dynamic_id = 1,
};

static struct class *lonusb_dev_class = 0;
