/*****************************************************************************/
/*
 * Linux driver for Easylon USB Interfaces
 *
 * Copyright (c)2003 Gesytec GmbH, Volker Schober (info@gesytec.de)
 * Licensend for use with Easylon USB Interfaces by Gesytec
 *
 * Please compile lonusb.c
 *
 */
 /*****************************************************************************/

typedef struct _read_element {
	struct list_head ListHead;
	size_t Len;
	u8 Data[USB_READBUFFER_SIZE];
} read_element;

NTSTATUS qwrite(PQUEUE queue, PUCHAR Data, ULONG Anz, BOOLEAN from_user)
{
	unsigned long flags;
	NTSTATUS status = 0;

	if (queue->MaxQueueCount > 0
	    && queue->QueueCount >= queue->MaxQueueCount)
		return -ENOMEM;

	read_element *re = kzalloc(sizeof(*re), GFP_ATOMIC);
	if (!re)
		return -ENOMEM;	// STATUS_INSUFFICIENT_RESOURCES;

	re->Len = Anz;
	if (from_user) {
		if (copy_from_user(re->Data, Data, Anz))
			status = -EFAULT;
	} else
		memmove(re->Data, Data, Anz);

	spin_lock_irqsave(&queue->polllock, flags);
	queue->QueueCount++;
	list_add_tail(&re->ListHead, &queue->read_list);
	spin_unlock_irqrestore(&queue->polllock, flags);

	return status;		// STATUS_SUCCESS;
}

LONG qread(PQUEUE queue, PUCHAR Data, ULONG MaxAnz, BOOLEAN to_user)
{
	unsigned long flags;
	read_element *re;

	spin_lock_irqsave(&queue->polllock, flags);

	re = list_first_entry(&queue->read_list, read_element, ListHead);
	list_del(&re->ListHead);
	queue->QueueCount--;

	spin_unlock_irqrestore(&queue->polllock, flags);

	{
		LONG Len = min((ULONG) re->Len, MaxAnz);
		if (Data && Len) {
			if (to_user) {
				if (copy_to_user(Data, re->Data, Len))
					Len = -EFAULT;
			} else
				memmove(Data, re->Data, Len);
		}
		kfree(re);

		return Len;
	}
}

BOOLEAN qempty(PQUEUE queue)
{
	return list_empty(&queue->read_list);
}

//BOOLEAN qfilled(PQUEUE queue)
BOOLEAN qhasdata(PQUEUE queue)
{
	return !list_empty(&queue->read_list);
}

int qcount(PQUEUE queue)
{
	return queue->QueueCount;
}

VOID qinit(PQUEUE queue, int MaxQueueCount)
{
	INIT_LIST_HEAD(&queue->read_list);
	spin_lock_init(&queue->polllock);
	queue->MaxQueueCount = MaxQueueCount;
}

VOID qexit(PQUEUE queue)
{
	qflush(queue);
//    DeleteCriticalSection(&queue->polllock);
}

NTSTATUS qflush(PQUEUE queue)
{
	read_element *re, *tmp;
	unsigned long flags;

	spin_lock_irqsave(&queue->polllock, flags);
	list_for_each_entry_safe(re, tmp, &queue->read_list, ListHead) {
		list_del(&re->ListHead);
		queue->QueueCount--;
		kfree(re);
	}
	spin_unlock_irqrestore(&queue->polllock, flags);

	return 0;		// STATUS_SUCCESS;
}
