#ifndef _NEURON_H
#define _NEURON_H

/****************************************************************************
 *  Filename:     neuron.h
 *
 *  Description:  This include file contains Neuron data structure definitions.
 *  It is equivalent to the portable versions of the following LonBuilder 3.0
 *  files:
 *      - addrdefs.h
 *      - access.h
 *      - status.h (with the addition of macros for evaluating reset cause)
 *      - msg_addr.h
 *
 *  Copyright (c) 1994-1997 Echelon Corporation.  All Rights Reserved.
 *
 *  Modification History:
 *    Name      Date        Description
 *
 ***************************************************************************/

#ifdef _ECHELON			/* compiled by Neuron-C compiler */
#ifndef USE_NEURON_NATIVE_TYPES
#define USE_NEURON_NATIVE_TYPES
#endif
#endif

#include "ni_platform.h"

#ifdef _MSC_VER
#pragma pack(_NSS_PACKING)
#endif

#if defined LINUX

#pragma pack(1)

#endif

#define NV_NAME_LEN         16	/* maximum base name length for names of NVs */
#define LMOBJ_NAME_LEN      16	/* maximum base name length for names of LonMark objects */
#define NS_ARRAY_SUB_SIZE   6	/* size of brackets and array index inside a */
				    /* variable name. */
#define NV_FULL_NAME_LEN    NV_NAME_LEN+NS_ARRAY_SUB_SIZE+1
				    /* maximum name length for names of NVs */
				    /* including array index and terminator. */
#define LMOBJ_FULL_NAME_LEN LMOBJ_NAME_LEN+NS_ARRAY_SUB_SIZE+1
				    /* maximum name length for names of NVs */
				    /* including array index and terminator. */
/*********************************************************************
   NUM_ADDR_TBL_ENTRIES:
   The address table in reference implementation supports more than
   15 entries. However, network management tools might support only
   upto 15 entries. So, other entries should be managed by the
   application program itself. The maximum supported value is 255.
**********************************************************************/
#if defined ADDRESS_EXTENSION
#define NUM_ADDR_TBL_ENTRIES 4096	/* # of entries in addr tbl    */
#else
#define NUM_ADDR_TBL_ENTRIES   15	/* # of entries in addr tbl    */
#endif

#define NUM_ADDR_TBL_ENTRIES_NEURON   15	/* # of entries in addr tbl    */

#define NULL_ADDR_INDEX     15	/* no address table index specified */
#define NULL_NV_SELECTOR    0xFFFF	/* indicates no NV selector used */
#define MAX_NEURON_NV_COUNT 62	/* Max # of NV definitions allowed for Neuron based nodes */
#define MAX_NV_COUNT        4096	/* Maximum NVs allowed for host nodes */
#define MAX_MSG_TAG_COUNT   15	/* Maximum # of message definitions allowed */
#define MAX_NEURON_ALIAS_COUNT 62	/* Max # of alias definitions allowed for Neuron based nodes */
#define MAX_ALIAS_COUNT     8192	/* Max # of alias definitions allowed for host based nodes */
#define MAX_NV_INDEX    (MAX_NV_COUNT-1)	/* Maximum NV index allowed for host nodes */
#define MAX_UNBOUND_SELECTOR 0x3FFF	/* largest unbound NV selector value */
#define MAX_BOUND_SELECTOR  0x2FFF	/* largest bound NV selector value */
#define NODE_OFFLINE_RESP   0x3F	/* Auto app response if node is offline */
#define NODE_FGN_OFFLINE_RESP 0x4F	/* Auto foreign resp if node is offline */

/*
 ****************************************************************************
 *  ABSOLUTE addresses
 ****************************************************************************
 */

#define OFS_FW_VERSION  0x0000

#define OFS_STAT_3150_V4   0xE918	/* for V4 and below */
#define OFS_STAT_3150_V7   0xE904	/* for V7, V11 and ???   */
#define OFS_STAT_3120_V3   0xED18	/* for V3 and below */

#define OFS_EXTRA_STATS    0x000A	/* offset of extra statistics */

#define ABS_READ_ONLY      0xF000	// READ_ONLY_RELATIVE
#define ABS_CONFIG         0xF029	// CONFIG_RELATIVE
#define ABS_DOMAIN_TABLE   0xF042	//

#define ABS_ERROR_LOG      0xF1FD	//
#define BOOT_ID_ADDR       0xF1FE

/* Boot ID enums, used with NSS10 Rel 1.02 and earlier firmware */
#define NSS_10_VER_1_00A_BOOT_ID 0x0154
#define NSS_10_VER_1_00B_BOOT_ID 0x0125
#define NSS_10_VER_1_01_BOOT_ID  0x0101
#define NSS_10_VER_1_02_BOOT_ID  0x0102

/*
 ****************************************************************************
 * CONFIG structure addresses
 ****************************************************************************
 */

#define LOCATION_ID_LEN 6

typedef struct {
	NetWord chan_id;
	Byte location[LOCATION_ID_LEN];
} config_struct;

#define OFS_CONFIG_STRUCT 0x0000

#define NUM_SPMODE_PARAMS 7

typedef struct {
	bits input_clock:3;
	bits comm_clock:5;
	bits comm_pin_dir:5;
	bits comm_type:3;
	Byte preamble_length;
	Byte packet_cycle;
	Byte beta2_control;
	Byte xmit_interpacket;
	Byte recv_interpacket;
	Byte node_priority;
	Byte channel_priorities;
	union {
		Byte spmode_params[NUM_SPMODE_PARAMS];
		struct {
			bits hysteresis:3;
			bits filter:2;
			bits bit_sync_threshold:2;
			bits collision_detect:1;
			bits cd_preamble:1;
			bits cd_tail:1;
			bits cd_to_end_packet:6;
		} dir_params;
	} params;
} xcvr_param_type;

// 3120, 3150
typedef enum { CLOCK_312, CLOCK_625, CLOCK_1250, CLOCK_2500,
	CLOCK_5000, CLOCK_10000, CLOCK_20000, CLOCK_40000
} input_clock_type;

// 5000
typedef enum { CLOCK5_40000, CLOCK5_80000, CLOCK5_625, CLOCK5_1250,
	CLOCK5_2500, CLOCK5_5000, CLOCK5_10000, CLOCK5_20000
} input_clock_type_5000;

typedef enum {
	COMM_TYPE_BLANK_NEURON = 0,
	COMM_TYPE_SINGLE_ENDED = 1,
	COMM_TYPE_DIFFERENTIAL = 5,
	COMM_TYPE_SPECIAL_PURPOSE = 2
} comm_type_type;

typedef enum {
	COMM_PIN_DIR_BLANK_NEURON = 0,
	COMM_PIN_DIR_SINGLE_ENDED = 0x0E,
	COMM_PIN_DIR_DIFFERENTIAL = 0x0C,
	COMM_PIN_DIR_SPECIAL_PURPOSE_WAKEUP_IN = 0x17,
	COMM_PIN_DIR_SPECIAL_PURPOSE_WAKEUP_OUT = 0x1E,
} comm_pin_dir_type;

#define OFS_COMM_PARAMS         ( OFS_CONFIG_STRUCT + sizeof( config_struct ) )
#define OFS_COMM_TYPE           ( OFS_COMM_PARAMS + 1 )

#define OFS_CHANNEL_PRIORITIES  \
    ( OFS_COMM_PARAMS + offsetof( xcvr_param_type, channel_priorities ))

#define OFS_SP_PARAMS           \
    ( OFS_COMM_PARAMS + offsetof( xcvr_param_type, params.spmode_params ))

#define COMM_PARAM_BYTES 11

// #define COMM_PARAM_BYTES        ( OFS_SP_PARAMS + 2 - OFS_COMM_PARAMS )
    // include dir_params, or first two bytes of spmode params

typedef struct {
	bits preempt_timeout:3;
	bits nm_auth:1;
	bits non_group_timer:4;
} config_misc;

#define OFS_CONFIG_MISC         ( OFS_COMM_PARAMS + sizeof( xcvr_param_type ) )
#define OFS_RTR_ALGORITHM       0x0037	// offset of nd_rtr_algorithm

/****************************************************************************
 * This section is equivalent to the LonBuilder include file status.h
 ***************************************************************************/
/* #define _STATUS */
typedef struct status_struct {
#ifdef USE_NEURON_NATIVE_TYPES	/* Neuron-C code or Echelon internal use */
	NetWord status_xmit_errors;
	NetWord status_transaction_timeouts;
	NetWord status_rcv_transaction_full;
	NetWord status_lost_msgs;
	NetWord status_missed_msgs;
	Byte status_reset_cause;	/* use resetCause... macros below to eval */
	Byte status_node_state;
	Byte status_version_number;
	Byte status_error_log;
	Byte status_model_number;
#else				/* for NI and NS applications */
	NetWord xmit_errors;	// number of CRC errors during packet reception
	NetWord transaction_timeouts;
	NetWord rcv_transaction_full;
	NetWord lost_msgs;
	NetWord missed_msgs;
	Byte reset_cause;	/* use resetCause... macros below to eval */
	Byte node_state;
	Byte version_number;
	Byte error_log;
	Byte model_number;
#endif
} status_struct;

/* Reset cause */
typedef enum {
	POWER_UP_RESET = 0x01,
	EXTERNAL_RESET = 0x02,
	WATCHDOG_RESET = 0x0C,
	SOFTWARE_RESET = 0x14,
	CLEARED = 0x00
} ResetCause;

/* macros for evaluating the reset cause Byte.  Example usage:
 *
 *      status_struct status;
 *      ....
 *      if (resetCauseWDT(status.reset_cause))
 *      {
 *          ....
 *      }
 *      ...
 */
#define resetCauseCleared(cause)  (cause == 0)
#define resetCausePowerUp(cause)  ((cause & 1)    == 1)
#define resetCauseExternal(cause) ((cause & 3)    == 2)
#define resetCauseWDT(cause)      ((cause & 0xf)  == 0xc)
#define resetCauseSoftware(cause) ((cause & 0x1f) == 0x14)

/* yes: Yes, error is implemented in reference implementation */
/* na:  Not applicable (not implemented) in ref. imp.         */
typedef enum {
	NO_ERRORS = 0,		/* yes */
	BAD_EVENT = 129,	/* na  */
	NV_LENGTH_MISMATCH = 130,	/* yes */
	NV_MSG_TOO_SHORT = 131,	/* yes */
	EEPROM_WRITE_FAIL = 132,	/* na  */
	BAD_ADDRESS_TYPE = 133,	/* yes */
	PREEMPTION_MODE_TIMEOUT = 134,	/* na  */
	ALREADY_PREEMPTED = 135,	/* na  */
	SYNC_NV_UPDATE_LOST = 136,	/* na  */
	INVALID_RESP_ALLOC = 137,	/* na  */
	INVALID_DOMAIN = 138,	/* yes */
	READ_PAST_END_OF_MSG = 139,	/* na  */
	WRITE_PAST_END_OF_MSG = 140,	/* na  */
	INVALID_ADDR_TABLE_INDEX = 141,	/* yes */
	INCOMPLETE_MSG = 142,	/* na  */
	NV_UPDATE_ON_OUTPUT_NV = 143,	/* na  */
	NO_MSG_AVAIL = 144,	/* na  */
	ILLEGAL_SEND = 145,	/* na  */
	UNKNOWN_PDU = 146,	/* yes */
	INVALID_NV_INDEX = 147,	/* yes */
	DIVIDE_BY_ZERO = 148,	/* na  */
	INVALID_APPL_ERROR = 149,	/* na  */
	MEMORY_ALLOC_FAILURE = 150,	/* yes */
	WRITE_PAST_END_OF_NET_BUFFER = 151,	/* na  */
	APPL_CS_ERROR = 152,	/* na  */
	CNFG_CS_ERROR = 153,	/* na  */
	INVALID_XCVR_REG_ADDR = 154,	/* na  */
	XCVR_REG_TIMEOUT = 155,	/* na  */
	WRITE_PAST_END_OF_APPL_BUFFER = 156,	/* na  */
	IO_READY = 157,		/* na  */
	SELF_TEST_FAILED = 158,	/* na  */
	SUBNET_ROUTER = 159,	/* na  */
	AUTHENTICATION_MISMATCH = 160,	/* yes */
	SELF_INST_SEMAPHORE_SET = 161,	/* na  */
	READ_WRITE_SEMAPHORE_SET = 162,	/* na  */
	APPL_SIGNATURE_BAD = 163,	/* na  */
	ROUTER_FIRMWARE_VERSION_MISMATCH = 164,	/* na  */
	ERROR_165 = 165,
	EEPROM_RECOVERY = 166,
	TRIAC_CLOCKEDGE = 167,
	SYS_CS_ERROR = 168,
	INVALID_PROXY_INDEX = 169,
	APPL_FIRMWARE_VERSION_MISMATCH = 170,
	NASCENT_STATE = 171,
	TRANSCEIVER_VIOLATION = 172,
	POSSIBLE_DEADLOCK = 173,
	STACK_ERROR = 174,
	INVALID_INSTRUCTION = 175,
	ISR_ACCESS_VIOLATION = 176,
	BREAKPOINT_IN_ISR = 177,
	WRITE_TO_SYSTEMIMAGE = 178
} ErrorLog;

/****************************************************************************
 * This section is equivalent to the LonBuilder include file addredefs.h
 ***************************************************************************/
/* #define _ADDRDEFS */

typedef enum _addr_type {
	UNBOUND = 0,
	SUBNET_NODE,
	NEURON_ID,
	BROADCAST,
	MULTICAST,
	MULTICAST_ACK,
	BROADCAST_GROUP = 0x23
} _addr_type;
typedef NetEnum(_addr_type) addr_type;
#define BROADCAST_GROUP_FLAG    (0x20)

typedef struct group_struct {
#ifdef BITF_LITTLE_ENDIAN
	bits size:7;		/* group size (0 => huge group) */
	bits type:1;		/* 1 => group */

	bits member:7;		/* member num (if huge group, only use 0) */
	bits domain:1;		/* domain index */

	bits retry:4;		/* retry count */
	bits rpt_timer:4;	/* unackd_rpt timer */

	bits tx_timer:4;	/* transmit timer index */
	bits rcv_timer:4;	/* receive timer index */
#else
	bits type:1;		/* 1 => group */
	bits size:7;		/* group size (0 => huge group) */

	bits domain:1;		/* domain index */
	bits member:7;		/* member num (if huge group, only use 0) */

	bits rpt_timer:4;	/* unackd_rpt timer */
	bits retry:4;		/* retry count */

	bits rcv_timer:4;	/* receive timer index */
	bits tx_timer:4;	/* transmit timer index */
#endif
	Byte group;		/* group ID */
} group_struct;

#if defined ADDRESS_EXTENSION
typedef enum {
	GRP_NORMAL = 0,
	GRP_OUTPUT_ONLY = 1,
	GRP_INPUT_NO_ACK = 2
} group_restriction;
#endif

#if defined ADDRESS_EXTENSION
typedef struct group_struct_ext {
#ifdef BITF_LITTLE_ENDIAN
	bits size:7;		/* group size (0 => huge group) */
	bits type:1;		/* 1 => group */

	bits member:6;		/* member num (if huge group, only use 0) */
	bits restriction:2;	/* domain index */

	bits retry:4;		/* retry count */
	bits rpt_timer:4;	/* unackd_rpt timer */

	bits tx_timer:4;	/* transmit timer index */
	bits rcv_timer:4;	/* receive timer index */
#else
	bits type:1;		/* 1 => group */
	bits size:7;		/* group size (0 => huge group) */

	bits restriction:2;	/* domain index */
	bits member:6;		/* member num (if huge group, only use 0) */

	bits rpt_timer:4;	/* unackd_rpt timer */
	bits retry:4;		/* retry count */

	bits rcv_timer:4;	/* receive timer index */
	bits tx_timer:4;	/* transmit timer index */
#endif
	Byte group;		/* group ID */
	NetWord domain;
} group_struct_ext;
#endif

typedef struct snode_struct {
#ifdef BITF_LITTLE_ENDIAN
	bits type:6;		/* SUBNET_NODE */
	bits key_override:2;

	bits node:7;		/* node number */
	bits domain:1;		/* domain index */

	bits retry:4;		/* retry count */
	bits rpt_timer:4;	/* unackd_rpt timer */

	bits tx_timer:4;	/* transmit timer index */
	bits:4;
#else
	bits key_override:2;
	bits type:6;		/* SUBNET_NODE */

	bits domain:1;		/* domain index */
	bits node:7;		/* node number */

	bits rpt_timer:4;	/* unackd_rpt timer */
	bits retry:4;		/* retry count */

	bits:4;
	bits tx_timer:4;	/* transmit timer index */
#endif
	Byte subnet;		/* subnet ID */
} snode_struct;

#if defined ADDRESS_EXTENSION
typedef struct snode_struct_ext {
#ifdef BITF_LITTLE_ENDIAN
	bits type:6;		/* SUBNET_NODE */
	bits key_override:2;

	bits node:7;		/* node number */
	bits:1;			/* domain index */

	bits retry:4;		/* retry count */
	bits rpt_timer:4;	/* unackd_rpt timer */

	bits tx_timer:4;	/* transmit timer index */
	bits:4;
#else
	bits key_override:2;
	bits type:6;		/* SUBNET_NODE */

	bits:1;			/* domain index */
	bits node:7;		/* node number */

	bits rpt_timer:4;	/* unackd_rpt timer */
	bits retry:4;		/* retry count */

	bits:4;
	bits tx_timer:4;	/* transmit timer index */
#endif
	Byte subnet;		/* subnet ID */
	NetWord domain;
} snode_struct_ext;
#endif

typedef struct bcast_struct {
#ifdef BITF_LITTLE_ENDIAN
	bits type:6;		/* BROADCAST */
	bits key_override:2;

	bits backlog:6;		/* backlog override value */
	bits:1;
	bits domain:1;		/* domain index */

	bits retry:4;		/* retry count */
	bits rpt_timer:4;	/* unackd_rpt timer */

	bits tx_timer:4;	/* transmit timer index */
	bits max_responses:4;
#else
	bits key_override:2;
	bits type:6;		/* BROADCAST */

	bits domain:1;		/* domain index */
	bits:1;
	bits backlog:6;		/* backlog override value */

	bits rpt_timer:4;	/* unackd_rpt timer */
	bits retry:4;		/* retry count */

	bits max_responses:4;
	bits tx_timer:4;	/* transmit timer index */
#endif
	Byte subnet;		/* subnet ID (0 => domain broadcast) */
} bcast_struct;

#if defined ADDRESS_EXTENSION
typedef struct bcast_struct_ext {
#ifdef BITF_LITTLE_ENDIAN
	bits type:6;		/* BROADCAST */
	bits key_override:2;

	bits backlog:6;		/* backlog override value */
	bits:1;
	bits:1;			/* domain index */

	bits retry:4;		/* retry count */
	bits rpt_timer:4;	/* unackd_rpt timer */

	bits tx_timer:4;	/* transmit timer index */
	bits max_responses:4;
#else
	bits key_override:2;
	bits type:6;		/* BROADCAST */

	bits:1;			/* domain index */
	bits:1;
	bits backlog:6;		/* backlog override value */

	bits rpt_timer:4;	/* unackd_rpt timer */
	bits retry:4;		/* retry count */

	bits max_responses:4;
	bits tx_timer:4;	/* transmit timer index */
#endif
	Byte subnet;		/* subnet ID (0 => domain broadcast) */
	NetWord domain;
} bcast_struct_ext;
#endif

/****************************************************************************
 * This section is equivalent to the LonBuilder include file access.h
 ***************************************************************************/
/* #define _ACCESS */

#define NEURON_ID_LEN    6
#define AUTH_KEY_LEN     6
#define DOMAIN_ID_LEN    6	// maximum domain ID length (possible values: 0, 1, 3, 6)
#define ID_STR_LEN       8
#define	PROGRAM_ID_LEN   ID_STR_LEN	/* 8 bytes per program ID */
#define LOCATION_LEN     6
#define NUM_COMM_PARAMS  7

/* This address type is the 'turnaround', used for connections solely in a node. */

typedef struct turnaround_struct {
	addr_type type;		/* UNBOUND */
	Byte turnaround;	/* 1 => Turnaround entry */
#ifdef BITF_LITTLE_ENDIAN
	bits retry:4;		/* retry count */
	bits rpt_timer:4;	/* unackd_rpt timer */

	bits tx_timer:4;	/* transmit timer index */
	bits:4;
#else
	bits rpt_timer:4;	/* unackd_rpt timer */
	bits retry:4;		/* retry count */

	bits:4;
	bits tx_timer:4;	/* transmit timer index */
#endif
} turnaround_struct;

#if defined ADDRESS_EXTENSION
typedef turnaround_struct turnaround_struct_ext;
#endif

/* Typedef for the address_struct returned by access_address(). */

typedef union address_struct {
	group_struct gp;
	snode_struct sn;
	bcast_struct bc;
	turnaround_struct ta;
} address_struct;

#if defined ADDRESS_EXTENSION
typedef union address_struct_ext {
	group_struct_ext gp;
	snode_struct_ext sn;
	bcast_struct_ext bc;
	turnaround_struct_ext ta;
	struct {
		address_struct base_struct;
		NetWord domain;
	} common;
} address_struct_ext;
#endif

/* Typedef for direct-transceiver parameters (used in the config_data_struct) */

typedef struct direct_param_struct {
#ifdef BITF_LITTLE_ENDIAN
	bits hysteresis:3;
	bits filter:2;
	bits bit_sync_threshold:2;
	bits collision_detect:1;

	bits cd_preamble:1;
	bits cd_tail:1;
	bits cd_to_end_packet:6;
#else
	bits collision_detect:1;
	bits bit_sync_threshold:2;
	bits filter:2;
	bits hysteresis:3;

	bits cd_to_end_packet:6;
	bits cd_tail:1;
	bits cd_preamble:1;
#endif
} direct_param_struct;

/* Typedef for the config_data_struct: */

typedef struct config_data_struct {
	NetWord channel_id;
	char location[LOCATION_LEN];
#ifdef BITF_LITTLE_ENDIAN
	bits input_clock:3;
	bits comm_clock:5;

	bits comm_pin_dir:5;
	bits comm_type:3;
#else
	bits comm_clock:5;
	bits input_clock:3;

	bits comm_type:3;
	bits comm_pin_dir:5;
#endif
	Byte preamble_length;
	Byte packet_cycle;
	Byte beta2_control;
	Byte xmit_interpacket;
	Byte recv_interpacket;
	Byte node_priority;
	Byte channel_priorities;
	union {
		Byte xcvr_params[NUM_COMM_PARAMS];
		direct_param_struct dir_params;
	} params;
#ifdef BITF_LITTLE_ENDIAN
	bits preemption_timeout:3;
	bits nm_auth:1;
	bits non_group_timer:4;
#else
	bits non_group_timer:4;
	bits nm_auth:1;
	bits preemption_timeout:3;
#endif
} config_data_struct;

/* The nm_auth, non_group_timer and premption timeout Byte stored in the
   neuron.
*/
typedef struct NrnAuthNonGroupTimerData {
#ifdef LITTLE_ENDIAN
	BitField preemption_timeout:3;
	BitField nm_auth:1;
	BitField non_group_timer:4;
#else				// neuron
	BitField non_group_timer:4;
	BitField nm_auth:1;
	BitField preemption_timeout:3;
#endif
} NrnAuthNonGroupTimerData;

typedef struct nsa_data_struct {
	Byte nsa_flag;		/* Enables NSA features (NM auth override, recv broadcast msgs when not on 0-length domain) */
	Byte auth_key[2][AUTH_KEY_LEN];	/* Specifies optional alternate authentication for outgoing messages */
} nsa_data_struct;

/*
typedef struct domain_struct
{
    Byte    id [DOMAIN_ID_LEN];
    Byte    subnet;
#ifdef BITF_LITTLE_ENDIAN
    bits    node    : 7;
    bits    cloneDomain : 1;
#else
    bits    cloneDomain : 1;
    bits    node    : 7;
#endif
    Byte    len;
    Byte    key [AUTH_KEY_LEN];
} domain_struct;
*/

/* Typedef for the nv_struct returned by the function access_nv: */
typedef struct nv_struct {
#ifdef USE_NEURON_NATIVE_TYPES	/* Neuron-C code or Echelon internal use */
#ifdef BITF_LITTLE_ENDIAN
	bits nv_selector_hi:6;
	bits nv_direction:1;	/* 1 => output */
	bits nv_priority:1;

	Byte nv_selector_lo;

	bits nv_addr_index:4;	/* Address table index */
	bits nv_auth:1;
	bits nv_service:2;
	bits nv_turnaround:1;
#else
	bits nv_priority:1;
	bits nv_direction:1;	/* 1 => output */
	bits nv_selector_hi:6;

	Byte nv_selector_lo;

	bits nv_turnaround:1;
	bits nv_service:2;
	bits nv_auth:1;
	bits nv_addr_index:4;	/* Address table index */
#endif
#else				/* for NI and NS applications */
#ifdef BITF_LITTLE_ENDIAN
	bits selector_hi:6;
	bits direction:1;
	bits priority:1;

	bits selector_lo:8;

	bits addr_index:4;
	bits auth:1;
	bits service:2;
	bits turnaround:1;
#else
	bits priority:1;
	bits direction:1;
	bits selector_hi:6;

	bits selector_lo:8;

	bits turnaround:1;
	bits service:2;
	bits auth:1;
	bits addr_index:4;
#endif
#endif
} nv_struct;

#if defined ADDRESS_EXTENSION
typedef struct nv_struct_ext {
#ifdef BITF_LITTLE_ENDIAN
	bits selector_hi:6;
	bits direction:1;
	bits priority:1;

	bits selector_lo:8;

	bits _dummy1:4;
	bits remote_nm_auth:1;
	bits write_by_index:1;
	bits auth:1;
	bits turnaround:1;

	bits _dummy2:5;
	bits service:2;
	bits read_by_index:1;
#else
	bits priority:1;
	bits direction:1;
	bits selector_hi:6;

	bits selector_lo:8;

	bits turnaround:1;
	bits auth:1;
	bits write_by_index:1;
	bits remote_nm_auth:1;
	bits _dummy1:4;

	bits read_by_index:1;
	bits service:2;
	bits _dummy2:5;
#endif
	NetWord addr_index;
	NetWord target_index;
} nv_struct_ext;
#endif

/* Typedef for the alias_struct returned by the function access_alias: */

typedef struct alias_struct {
	nv_struct alias_nv;
	Byte alias_primary;
	Word host_primary;
} alias_struct;

#if defined ADDRESS_EXTENSION
typedef struct alias_struct_ext {
	nv_struct_ext alias_nv;
	NetWord alias_primary;
} alias_struct_ext;
#endif

/* Typedefs for the read_only_data_struct: */

typedef struct nv_fixed_struct {
#ifdef BITF_LITTLE_ENDIAN
	bits nv_length:5;	// # of bytes in the Network Var
	bits:2;			// unused
	bits nv_sync:1;		// 1 ==> Synchronous network var
#else
	bits nv_sync:1;		// 1 ==> Synchronous network var
	bits:2;			// unused
	bits nv_length:5;	// # of bytes in the Network Var
#endif
//    void *  nv_address;
	NeuronMemAddr nv_address;	// Ptr to variable's data or 0xFFFF
} nv_fixed_struct;

typedef struct nv_fixed_struct_internal {
#ifdef BITF_LITTLE_ENDIAN
	bits nv_length:5;	// # of bytes in the Network Var
	bits:2;			// unused
	bits nv_sync:1;		// 1 ==> Synchronous network var
#else
	bits nv_sync:1;		// 1 ==> Synchronous network var
	bits:2;			// unused
	bits nv_length:5;	// # of bytes in the Network Var
#endif
	NeuronMemAddr nv_address;	// Ptr to variable's data or 0xFFFF
	void *p_nv_address;	// real Ptr to variable's data in Mip-Emulation
} nv_fixed_struct_internal;

typedef struct snvt_struct {
	NetWord length;
	Byte num_netvars;	/* Max # of NVs */
	Byte version;
	union {
		struct {
			Byte mtag_count;
		} ver0;
		struct {
			Byte msb_num_netvars;
			Byte mtag_count;
		} ver1;
		struct {
			Byte msb_num_netvars;	/* Max # of NVs */
			Byte mtag_count;
			NetWord static_nv_count;
			NetWord current_nv_count;	/* Number of currently defined NVs */
			NetWord max_nv_in_use;	/* Maximum NV index */
			NetWord alias_count;
			NetWord node_sd_text_length;
#ifdef BITF_LITTLE_ENDIAN
			bits unused:6;
			bits query_stats:1;
			bits binding_II:1;
#else
			bits binding_II:1;
			bits query_stats:1;
			bits unused:6;
#endif
		} ver2;
	} variable_part;
	/* Further variable info is as follows: */
	/* 1 snvt_descriptor per NV */
	/* 1 NULL terminated string */
	/* 1 extension rec per NV with the ext_rec bit set */
} snvt_struct;

typedef struct snvt_capability_info {
	NetWord length;
	Byte ver_struct;
	Byte ver_nm_min;
	Byte ver_nm_max;
	Byte ver_binding;
	Byte flags[6];
	NetWord domain_capacity;
	NetWord address_capacity;
	NetWord static_mtag_capacity;
} snvt_capability_info;

#if defined ADDRESS_EXTENSION
typedef struct snvt_struct_v2 {
	NetWord length;
	Byte num_netvars;	/* Max # of NVs */
	Byte version;
	union {
		struct {
			Byte mtag_count;
		} ver0;
		struct {
			Byte msb_num_netvars;
			Byte mtag_count;
		} ver1;
		struct {
			Byte msb_num_netvars;	/* Max # of NVs */
			Byte mtag_count;
			NetWord static_nv_count;
			NetWord current_nv_count;	/* Number of currently defined NVs */
			NetWord max_nv_in_use;	/* Maximum NV index */
			NetWord alias_count;
			NetWord node_sd_text_length;
#ifdef BITF_LITTLE_ENDIAN
			bits unused:6;
			bits query_stats:1;
			bits binding_II:1;
#else
			bits binding_II:1;
			bits query_stats:1;
			bits unused:6;
#endif
		} ver2;
	} variable_part;

	snvt_capability_info cap;
} snvt_struct_v2;
#endif

#define SNVT_V0_HDR_SIZE    (sizeof(snvt_struct) - sizeof(((snvt_struct *)0)->variable_part) + sizeof(((snvt_struct *)0)->variable_part.ver0))
#define SNVT_V1_HDR_SIZE    (sizeof(snvt_struct) - sizeof(((snvt_struct *)0)->variable_part) + sizeof(((snvt_struct *)0)->variable_part.ver1))
#define SNVT_V2_HDR_SIZE    (sizeof(snvt_struct) - sizeof(((snvt_struct *)0)->variable_part) + sizeof(((snvt_struct *)0)->variable_part.ver2))

#define SI_VER_SUPPORTS_ACCESS_METHODS(ver)     ((ver) >= 2)
#define SI_VER_SUPPORTS_DYNAMIC_NV(ver)         SI_VER_SUPPORTS_ACCESS_METHODS(ver)

/* Values for snvtPtr */
#define SIDATA_NONE 0
#define SIDATA_IN_HOST ((NeuronMemAddr)0xffff)

/* structure of the SI data's NV description record, one per NV declaration */
typedef struct snvt_desc_struct {
#ifdef BITF_LITTLE_ENDIAN
	bits nv_config_class:1;
	bits nv_auth_config:1;
	bits nv_priority_config:1;
	bits nv_service_type_config:1;
	bits nv_offline:1;
	bits nv_polled:1;
	bits nv_sync:1;
	bits ext_rec:1;
#else
	bits ext_rec:1;
	bits nv_sync:1;
	bits nv_polled:1;
	bits nv_offline:1;
	bits nv_service_type_config:1;
	bits nv_priority_config:1;
	bits nv_auth_config:1;
	bits nv_config_class:1;
#endif
	Byte snvt_type_index;	/* use enum SNVT_t */
} snvt_desc_struct;

/* structure of the SI data's NV extension record, one per NV declaration if
   the its snvt_desc_struct.ext_rec is 1. */
typedef struct snvt_ext_rec_mask {
#ifdef BITF_LITTLE_ENDIAN
	bits unused:3;
	bits nc:1;		/* array count */
	bits sd:1;		/* self-documentation */
	bits nm:1;		/* network variable name */
	bits re:1;		/* rate estimate */
	bits mre:1;		/* max rate estimate */
#else
	bits mre:1;		/* max rate estimate */
	bits re:1;		/* rate estimate */
	bits nm:1;		/* network variable name */
	bits sd:1;		/* self-documentation */
	bits nc:1;		/* array count */
	bits unused:3;
#endif
} snvt_ext_rec_mask;

#define ALIAS_COUNT_ESCAPE      63

typedef struct snvt_alias_field {
#ifdef BITF_LITTLE_ENDIAN
	bits alias_count:6;
	bits query_stats:1;
	bits binding_II:1;
#else
	bits binding_II:1;
	bits query_stats:1;
	bits alias_count:6;
#endif
	NetWord host_alias_count;	/* Host alias count if alias_count is ALIAS_COUNT_ESCAPE (63) */
} snvt_alias_field;

typedef struct EncodedBuffers {
	Byte app_buf_sizes;
	Byte net_buf_sizes;
	Byte priority_buf_counts;
	Byte app_buf_counts;
	Byte net_buf_counts;
} EncodedBuffers;

typedef struct read_only_data_struct {
	Byte neuron_id[NEURON_ID_LEN];	// uniqueNodeId[UNIQUE_NODE_ID_LEN]
	Byte model_num;		// offset 06
#ifdef BITF_LITTLE_ENDIAN
	bits minor_model_num:4;	// offset 07
	bits nrnid_xsum:4;
	NeuronMemAddr nv_fixed;	// offset 08  ptr nv_fixed_struct
	bits nv_count:6;	// offset 0A
	bits run_if_unconfigd:1;
	bits read_write_protect:1;
	NeuronMemAddr snvt;	// offset 0B  ptr snvt_struct
	Byte id_string[ID_STR_LEN];	// offset 0D  progId
	bits node_state:3;	// offset 15  state
	bits msg_process:1;
	bits no_sched_reset:1;
	bits explicit_addr:1;	// expl_addr
	bits two_domains:1;
	bits NV_processing_off:1;	// net_funct

	bits num_io_chgs:4;
	bits address_count:4;	// num_addr

	bits receive_trans_count:4;	// offset 17
	bits timer_count:4;

	bits app_buf_in_size:4;
	bits app_buf_out_size:4;

	bits net_buf_in_size:4;
	bits net_buf_out_size:4;

	bits app_buf_out_priority_count:4;
	bits net_buf_out_priority_count:4;

	bits app_buf_in_count:4;
	bits app_buf_out_count:4;

	bits net_buf_in_count:4;
	bits net_buf_out_count:4;
//    Byte    reserved1 [6];                      // offset 1D
//    bits    idempotent_duplicate    : 1;        // offset 23
//    bits    tx_by_address           : 1;
//    bits                            : 6;
	Byte eeprom_count;	// offset 1D
	Byte msg_compl_lo;	// offset 1E
	bits msg_compl_hi:7;	// offset 1F
	bits nv_succeeds:1;

	Byte io_dir_0_7;	// direction bits 0-7  // offset 20
	bits io_data_8_10:3;	// offset 21
	bits io_dir_8_10:3;
	bits io_parallel:1;
	bits io_master:1;

	Byte io_data_0_7;	// offset 22
	bits idem_dupl:1;	// offset 23  idempotent_duplicate
	bits tx_by_addr:1;	// tx_by_address
	bits io_pullups:1;
	bits tc1_out_ena:1;
	bits tc2_out_ena:1;
	bits io_tc_mux:2;
#else
	bits nrnid_xsum:4;	// offset 07
	bits minor_model_num:4;
	NeuronMemAddr nv_fixed;	// offset 08  ptr nv_fixed_struct
	bits read_write_protect:1;	// offset 0A
	bits run_if_unconfigd:1;
	bits nv_count:6;
	NeuronMemAddr snvt;	// offset 0B  ptr snvt_struct */
	Byte id_string[ID_STR_LEN];	// offset 0D
	bits NV_processing_off:1;	// offset 15  net_funct
	bits two_domains:1;
	bits explicit_addr:1;	// expl_addr
	bits no_sched_reset:1;
	bits msg_process:1;
	bits node_state:3;	// state

	bits address_count:4;	// num_addr
	bits num_io_chgs:4;

	bits timer_count:4;	// offset 17
	bits receive_trans_count:4;

	bits app_buf_out_size:4;
	bits app_buf_in_size:4;

	bits net_buf_out_size:4;
	bits net_buf_in_size:4;

	bits net_buf_out_priority_count:4;
	bits app_buf_out_priority_count:4;

	bits app_buf_out_count:4;
	bits app_buf_in_count:4;

	bits net_buf_out_count:4;
	bits net_buf_in_count:4;
//    Byte    reserved1 [6];                      // offset 1D
//    bits                            : 6;        // offset 23
//    bits    tx_by_address           : 1;
//    bits    idempotent_duplicate    : 1;
	Byte eeprom_count;	// offset 1D
	Byte msg_compl_lo;	// offset 1E
	bits nv_succeeds:1;	// offset 1F
	bits msg_compl_hi:7;

	Byte io_dir_0_7;	// direction bits 0-7  // offset 20
	bits io_master:1;	// offset 21
	bits io_parallel:1;
	bits io_dir_8_10:3;
	bits io_data_8_10:3;

	Byte io_data_0_7;	// offset 22
	bits io_tc_mux:2;	// offset 23
	bits tc2_out_ena:1;
	bits tc1_out_ena:1;
	bits io_pullups:1;
	bits tx_by_addr:1;	// tx_by_address
	bits idem_dupl:1;	// idempotent_duplicate
#endif
} read_only_data_struct;

/**************************************************************************

    This data structure is a subset of the read_only_data_struct and
    describes the buffer configuration in the neuron.

   ***********************************************************************/
typedef struct neuron_buffer_config {
#ifdef LITTLE_ENDIAN
	bits receive_trans_count:4;
	bits timer_count:4;

	bits app_buf_in_size:4;
	bits app_buf_out_size:4;

	bits net_buf_in_size:4;
	bits net_buf_out_size:4;

	bits app_buf_out_priority_count:4;
	bits net_buf_out_priority_count:4;

	bits app_buf_in_count:4;
	bits app_buf_out_count:4;

	bits net_buf_in_count:4;
	bits net_buf_out_count:4;
#else
	bits timer_count:4;
	bits receive_trans_count:4;

	bits app_buf_out_size:4;
	bits app_buf_in_size:4;

	bits net_buf_out_size:4;
	bits net_buf_in_size:4;

	bits net_buf_out_priority_count:4;
	bits app_buf_out_priority_count:4;

	bits app_buf_out_count:4;
	bits app_buf_in_count:4;

	bits net_buf_out_count:4;
	bits net_buf_in_count:4;
#endif
} neuron_buffer_config;
//#define OFFSETOF_BUFFER_CONFIG (offsetof(read_only_data_struct, reserved1) - sizeof(neuron_buffer_config))
#define OFFSETOF_BUFFER_CONFIG (offsetof(read_only_data_struct, eeprom_count) - sizeof(neuron_buffer_config))

/* The following addendum to the read-only data struct is available */
/* only in firmware version 6 and later. */
typedef struct read_only_data_struct_2 {
#ifdef BITF_LITTLE_ENDIAN
	bits alias_count:6;	// offset 24
	bits:2;

	bits:3;
	bits capability_info:1;
	bits msg_tag_count:4;
#else
	bits:2;			// offset 24
	bits alias_count:6;

	bits msg_tag_count:4;
	bits capability_info:1;
	bits:3;
#endif
	Byte larva_entry[3];	// offset 26
} read_only_data_struct_2;

/* The following externs are use only when compiling with neuron-C */
#ifdef _ECHELON
/* Read only data structures */

extern system const config_data_struct config_data;

extern system const read_only_data_struct read_only_data;

extern system const read_only_data_struct_2 read_only_data_2;

/* Access functions which return pointers to const (read-only) data */

extern system far const address_struct *access_address(int index);

extern system far const domain_struct *access_domain(int index);

extern system far const nv_struct *access_nv(int index);

extern system far const alias_struct *access_alias(int index);

/* Self-configuration functions used to modify own config data */

extern system far void update_address(const address_struct * address,
				      int index);

extern system far void update_domain(domain_struct * domain, int index);

extern system far void update_clone_domain(domain_struct * domain, int index);

extern system far void update_nv(const nv_struct * nv_entry, int index);

extern system far void update_alias(const alias_struct * alias_entry,
				    int index);

extern system far void update_config_data(const config_data_struct *
					  config_data);

	/* Was last reset due to powerup? */
extern system far boolean power_up(void);

	/* Is the system currently in preemption mode? */
extern system far boolean preemption_mode(void);

	/* Fill in the system status struct passed in by pointer */
extern system far void retrieve_status(status_struct * status_p);

	/* Clear the system status values */
extern system far void clear_status(void);

	/* Return the transceiver status from reg 1..7 */
extern system unsigned long retrieve_xcvr_status(unsigned reg);
#endif

/****************************************************************************
 * This section is equivalent to the LonBuilder include file msg_addr.h
 * It contains definitions for structures used with the Neuron C
 * messaging facility's explicit addressing feature.
 ***************************************************************************/
/* #define _MSG_ADDR */

/* Typedefs for 'msg_out_addr', the type of the field 'msg_out.dest_addr'. */

typedef struct nrnid_struct {
#ifdef BITF_LITTLE_ENDIAN
	bits type:6;		/* NEURON_ID */
	bits key_override:2;

	bits:7;
	bits domain:1;

	bits retry:4;
	bits rpt_timer:4;

	bits tx_timer:4;
	bits:4;
#else
	bits key_override:2;
	bits type:6;		/* NEURON_ID */

	bits domain:1;
	bits:7;

	bits rpt_timer:4;
	bits retry:4;

	bits:4;
	bits tx_timer:4;
#endif
	Byte subnet;
	Byte nid[NEURON_ID_LEN];
} nrnid_struct;

#if defined ADDRESS_EXTENSION
typedef struct nrnid_struct_ext {
#ifdef BITF_LITTLE_ENDIAN
	bits type:6;		/* NEURON_ID */
	bits key_override:2;

	bits:7;
	bits:1;

	bits retry:4;
	bits rpt_timer:4;

	bits tx_timer:4;
	bits:4;
#else
	bits key_override:2;
	bits type:6;		/* NEURON_ID */

	bits:1;
	bits:7;

	bits rpt_timer:4;
	bits retry:4;

	bits:4;
	bits tx_timer:4;
#endif
	Byte subnet;
	Byte nid[NEURON_ID_LEN];
	Word domain;
} nrnid_struct_ext;
#endif

#ifndef ADDRESS_EXTENSION
typedef union msg_out_addr {
	addr_type no_address;	/* UNBOUND (i.e. 0) if no address */
	/* UNBOUND & addr_type are defined in <ADDRDEFS.H> */
	group_struct group;
	snode_struct snode;
	nrnid_struct nrnid;
	bcast_struct bcast;
} msg_out_addr;
#else
typedef union msg_out_addr {
	addr_type no_address;	/* UNBOUND (i.e. 0) if no address */
	/* UNBOUND & addr_type are defined in <ADDRDEFS.H> */
	group_struct_ext group;
	snode_struct_ext snode;
	nrnid_struct_ext nrnid;
	bcast_struct_ext bcast;
} msg_out_addr;
#endif
/* Typedef for 'msg_in_addr', which is the type of the field 'msg_in.addr'. */

typedef struct msg_in_addr {
#ifdef BITF_LITTLE_ENDIAN
	bits format:6;		/* 'RcvDstAddrFormat' enum. */
	bits flex_domain:1;
	bits domain:1;
	struct {
		Byte subnet;
		bits node:7;
		bits:1;
	} src_addr;
	union {
		Byte bcast_subnet;
		Byte group;
		struct {
			Byte subnet;
			bits node:7;
			bits:1;
		} snode;
		struct {
			Byte subnet;
			Byte nid[NEURON_ID_LEN];
		} nrnid;
	} dest_addr;
#else
	bits domain:1;
	bits flex_domain:1;
	bits format:6;		/* 'RcvDstAddrFormat' enum. */
	struct {
		Byte subnet;
		bits:1;
		bits node:7;
	} src_addr;
	union {
		Byte bcast_subnet;
		Byte group;
		struct {
			Byte subnet;
			bits:1;
			bits node:7;
		} snode;
		struct {
			Byte subnet;
			Byte nid[NEURON_ID_LEN];
		} nrnid;
	} dest_addr;
#endif
} msg_in_addr;

/* Typedef for 'resp_in_addr', which is the type of the field 'resp_in.addr'. */

typedef struct resp_in_addr {
#ifdef BITF_LITTLE_ENDIAN
	bits format:6;		/* 'RcvDstAddrFormat' enum. */
	bits flex_domain:1;
	bits domain:1;
	struct {
		Byte subnet;
		bits node:7;
		bits is_snode:1;	/* 0=>group resp, 1=>snode resp */
	} src_addr;
	union {
		struct {
			Byte subnet;
			bits node:7;
			bits:1;
		} snode;
		struct {
			Byte subnet;
			bits node:7;
			bits:1;
			Byte group;
			bits member:6;
			bits:2;
		} group;
	} dest_addr;
#else
	bits domain:1;
	bits flex_domain:1;
	bits format:6;		/* 'RcvDstAddrFormat' enum. */
	struct {
		Byte subnet;
		bits is_snode:1;	/* 0=>group resp, 1=>snode resp */
		bits node:7;
	} src_addr;
	union {
		struct {
			Byte subnet;
			bits:1;
			bits node:7;
		} snode;
		struct {
			Byte subnet;
			bits:1;
			bits node:7;
			bits group;
			bits:2;
			bits member:6;
		} group;
	} dest_addr;
#endif
} resp_in_addr;

#ifdef _ECHELON
/* Prototypes for compiler's access functions for the fields */
/* 'msg_out.dest_addr', 'msg_in.addr', and 'resp_in.addr'. */
/* The programmer should NOT access these functions directly. */

extern system far msg_out_addr *_msg_out_addr_ptr(void);

extern system far msg_in_addr *_msg_in_addr_ptr(void);

extern system far const resp_in_addr *_resp_in_addr_ptr(void);
#endif

#ifdef _MSC_VER
#pragma pack()
#endif

#if defined LINUX
#pragma pack()
#endif

#endif
