#ifndef _NI_MGMT_H
#define _NI_MGMT_H

/****************************************************************************
 *  Filename:     ni_mgmt.h
 *
 *  Description:  This include file contains LonTalk network management
 *                codes and messages.  These definitions are a subset
 *                of the LonTalk network management structures.  The
 *                types have been changed where necessary because of the
 *                difference in representation between the Neuron and
 *                MS-DOS.  If you aren't using DOS you may need to modify
 *                these structures.
 *
 *  Copyright (c) 1994-1997 Echelon Corporation.  All Rights Reserved.
 *
 ***************************************************************************/

#ifdef _MSC_VER
#pragma pack(_NSS_PACKING)
#endif

#if defined LINUX
#pragma pack(1)
#endif

#ifdef __cplusplus
extern "C" {			/* Assume C declarations for C++ */
#endif				/* __cplusplus */

#define NULL_IDX  NULL_ADDR_INDEX	/* unused address table index */

#define ND_query_status          0x51
#define ND_query_status_fail     0x11
#define ND_query_status_succ     0x31

#define NM_update_nv_cnfg        0x6B
#define NM_update_nv_cnfg_fail   0x0B
#define NM_update_nv_cnfg_succ   0x2B

	typedef struct {
		Byte code;
		nv_struct nv_cnfg;
	} NM_query_nv_cnfg_response;

#define NM_query_nv_cnfg        0x68
#define NM_query_nv_cnfg_fail   0x08
#define NM_query_nv_cnfg_succ   0x28

	typedef enum {
		APPL_OFFLINE = 0,	/* Soft offline state */
		APPL_ONLINE,
		APPL_RESET,
		CHANGE_STATE,
		DEFAULT_RESET
	} nm_node_mode;

/*
   The neuron "state" reported by the query status command contains a 3 bit
   state value stored in the node's EEPROM, plus an online bit and an
   offline/bypass mode bit.  In general the offline and offline/bypass mode
   bits are useful only if the 3 bit neuron state state is "configured".
   However, for debugging purposes it might be useful to know if the node was
   taken offline prior to having its state changed - the offline bit indicates
   this (although if the node is reset after the state change the offline bit
   will not be set).

   The following literals and macros can be used to interpret the node state.

*/

#define OFFLINE_BIT 0x8
#define NODE_STATE_MASK 0x7
#define BYPASS_BIT 0x80

/* return the 3 bit neuron state stored in EEPROM. */
#define NEURON_STATE(state) ((state)&NODE_STATE_MASK)

/* return nm_node_state - literals defined below */
#define NODE_STATE(state) ((NEURON_STATE(state) == CNFG_ONLINE) \
    ? (state) : NEURON_STATE(state))

/* Return 0 if the state indicates offline, non-zero otherwise. */
#define NODE_STATE_OFFLINE(state) ((state) & OFFLINE_BIT)

/* This enumeration is used with
        NM_set_node_mode_request  and
        ND_query_status_response.

   This ENUM contains the documented values of the neuron state. To convert
   the state value returned by the ND_query_status command to one of these
   values, use the macro NODE_STATE() defined above.  To see if the offline
   bit is set, use the macro NODE_STATE_OFFLINE().
    */

	typedef enum {
		BLANK = 0,	/* invalid or echelon use only          */
		INIT_CS_UNCFG = 1,	/* equivalent to STATE_INVALID          */
		APPL_UNCNFG = 2,	/* has application, unconfigured        */
		NO_APPL_UNCNFG = 3,	/* applicationless, unconfigured        */
		CNFG_ONLINE = 4,	/* configured, online                   */
		INIT_CS_CNFG = 5,	/* equivalent to STATE_INVALID          */
		CNFG_OFFLINE = 6,	/* Configured, hard offline             */
		STATE_INVALID_7 = 7,	/* equivalent to STATE_INVALID          */
		SOFT_OFFLINE = (CNFG_ONLINE | OFFLINE_BIT),	/* (12) configured, soft-offline        */
		CNFG_BYPASS =	/* 0x8c - configured, in bypass mode    */
#ifdef COMPILE_8		/* special casting for neuron C compiler */
		    (signed char)
#endif
		    (BYPASS_BIT | SOFT_OFFLINE),	/* 0x8C, 140 (decimal)   */
	} nm_node_state;

	typedef struct {
		Byte code;
		Byte mode;	/* Interpret with 'nm_node_mode'        */
		Byte node_state;	/* Optional field if mode==CHANGE_STATE */
		/* Interpret with 'nm_node_state'       */
	} NM_set_node_mode_request;

#define NM_set_node_mode      0x6C
#define NM_set_node_mode_fail 0x0C
#define NM_set_node_mode_succ 0x2C

	typedef enum {
		ABSOLUTE_MEM_MODE = 0,	/* Address is absolute neuron mem address */
		READ_ONLY_RELATIVE = 1,	/* Address is offset from beginning of
					   read-only memory structures. */
		CONFIG_RELATIVE = 2,	/* Address is offset from beginning of
					   config data structures. */
		STATS_RELATIVE = 3,	/* Address is offset from beginning of
					   statistics data structures. */
		MEM_MODE_RESERVED_A = 4	/* Reserved for Echelon internal use only */
	} nm_mem_mode;

	typedef enum {
		NO_ACTION = 0,
		BOTH_CS_RECALC = 1,
		DELTA_CS_RECALC = 3,
		CNFG_CS_RECALC = 4,
		ONLY_RESET = 8,
		BOTH_CS_RECALC_RESET = 9,
		CNFG_CS_RECALC_RESET = 12,
		FLASH_SYSTEM_IMAGE = 0x80
	} nm_mem_form;

	typedef struct {
		Byte code;
		Byte mode;
		Byte offset_hi;
		Byte offset_lo;
		Byte count;
		Byte form;	/* followed by the data */
	} NM_write_memory_request;

/* this is an alternative declaration for NM_write_memory_request that
   resembles the declaration in the neuron-c include file */
	typedef struct {
		Byte code;
		Byte mode;
		NetWord offset;
		Byte count;
		Byte form;	/* followed by the data */
	} NM_write_memory_request_nc;

#define NM_write_memory             0x6E
#define NM_write_memory_fail        0x0E
#define NM_write_memory_success     0x2E

/* Partial list of SNVT type index values */

	typedef enum {
		SNVT_str_asc = 36,
		SNVT_lev_cont = 21,
		SNVT_lev_disc = 22,
		SNVT_count_f = 51
	} SNVT_t;

/* Network management message codes */

#define NM_query_SNVT           0x72
#define NM_query_SNVT_fail      0x12
#define NM_query_SNVT_succ      0x32

	typedef struct {
		Byte code;
		NetWord offset;	/* big-endian 16-bits */
		Byte count;
	} NM_query_SNVT_request;

#define NM_wink 0x70
#define NM_wink_succ                0x30
#define NM_wink_fail                0x10

	typedef enum {
		WINK = 0,
		SEND_ID_INFO = 1,
	} nm_install;

	typedef struct NM_install_request {
		nm_install subcommand;
		Byte network_interface;	// present if subcommand == SEND_ID_INFO
	} NM_install_request;

	typedef struct NM_install_response {	// resp to SEND_IF_INFO from host node
		Byte interface_down;
		Byte neuron_id[NEURON_ID_LEN];
		Byte id_string[ID_STR_LEN];
	} NM_install_response;

#define NM_NV_fetch         0x73
#define NM_NV_fetch_fail    0x13
#define NM_NV_fetch_succ    0x33

/* Structure used by host application to store network variables */

	typedef enum {
		NV_IN = 0,
		NV_OUT = 1
	} nv_direction;

	typedef struct {	/* structure to define NVs */
		int size;	/* number of Bytes */
		nv_direction direction;	/* input or output */
		const char *name;	/* name of variable */
		void (*print_func) (Byte *);	/* routine to print value */
		void (*read_func) (Byte *);	/* routine to read value */
		Byte data[MAX_NETVAR_DATA];	/* actual storage for value */
	} network_variable;

	typedef struct {
		Byte code;
		Byte index;
	} short_index_struct;

	typedef struct {
		short_index_struct short_index;
		nv_struct nv_cnfg_data;
	} short_config_data_struct;

	typedef struct {
		Byte code;
		Byte index_escape;	/* Must be 255 */
		NetWord index;
	} long_index_struct;

	typedef struct {
		long_index_struct index;
		nv_struct nv_cnfg_data;
	} long_config_data_struct;

	typedef struct {
		short_index_struct short_index;
		alias_struct alias_cnfg_data;
	} short_alias_config_struct;

	typedef struct {
		long_index_struct long_index;
		alias_struct alias_cnfg_data;
		NetWord host_primary_index;
	} long_alias_config_struct;

/* Update Network Variable Configuration Request
 */
	typedef union NM_update_nv_cnfg_request {
		short_config_data_struct short_config_index;
		long_config_data_struct long_config_index;
		short_alias_config_struct short_alias_config_index;
		long_alias_config_struct long_alias_config_index;
	} NM_update_nv_cnfg_request;

/* Query Network Variable Configuration Request
 */
	typedef union NM_query_nv_cnfg_request {
		short_index_struct short_index;
		long_index_struct long_index;
	} NM_query_nv_cnfg_request;

/*  Network Variable Fetch Request
 */
	typedef NM_query_nv_cnfg_request NM_nv_fetch_request;

/*  Network Variable Fetch Response
 */
	typedef union NM_nv_fetch_response {
		short_index_struct short_index;
		long_index_struct long_index;	/* followed by data */
	} NM_nv_fetch_response;

/* -----------------------------------------------------------------------
   The following data structures are used by hosts that are attached to
   a MIP rather than an NSS.  NSS-hosts do not need to use these data
   structures.
 */

/* Update Domain Request
 */
#define NM_update_domain            0x63
#define NM_update_domain_fail       0x03
#define NM_update_domain_success    0x23

	typedef struct {
		Byte code;
		Byte domain_index;
		Byte id[DOMAIN_ID_LEN];
		Byte subnet;
#ifdef BITF_LITTLE_ENDIAN
		bits node:7;
		bits cloneDomain:1;	// this bit must be set to 1
#else
		bits cloneDomain:1;	// this bit must be set to 1
		bits node:7;
#endif
		Byte len;
		Byte key[AUTH_KEY_LEN];
	} NM_update_domain_request;

/* Query Domain Request/Response
 */
#define NM_query_domain            0x6a
#define NM_query_domain_fail       0x0a
#define NM_query_domain_success    0x2a

	typedef struct {
		Byte code;
		Byte domain_index;
	} NM_query_domain_request;

	typedef struct {
		Byte code;
		Byte id[DOMAIN_ID_LEN];
		Byte subnet;
#ifdef BITF_LITTLE_ENDIAN
		bits node:7;
		bits cloneDomain:1;	// this bit must be set to 1
#else
		bits cloneDomain:1;	// this bit must be set to 1
		bits node:7;
#endif
		Byte len;
		Byte key[AUTH_KEY_LEN];
	} NM_query_domain_response;

/* Leave Domain Request
 */
#define NM_leave_domain            0x64
#define NM_leave_domain_fail       0x04
#define NM_leave_domain_success    0x24

	typedef struct {
		Byte code;
		Byte domain_index;
	} NM_leave_domain_request;

/* Memory Read Request
 */
#define NM_read_memory              0x6D
#define NM_read_memory_fail         0x0D
#define NM_read_memory_success      0x2D

	typedef struct {
		Byte code;
		Byte mode;	/* enum nm_mem_mode */
		NetWord offset;
		Byte count;
	} NM_read_memory_request;

/* Device Escape Command
 * The LONTalk network management commands for device escape all begin
 * with the following 3-byte sequence:
 *      <Device Escape Code> <Device Code> <Optional Command code>
 */
#define DEVICE_ESC_SEQ_LEN          3

#define NM_device_esc             0x7d

/* the following are not needed if you have real NSI MIP
   Delete this section after everyone has been upgraded.
#ifdef INCLUDE_NSI_ONLY
  #ifdef NEURON_HOSTED
    #undef NM_device_esc
    #define NM_device_esc              0x0A
  #endif
#endif
*/

#define NM_device_esc_fail          0x1d
#define NM_device_esc_success       0x3d

#define NM_ESCAPE_CODE NM_device_esc
/*
 * Header structure for all messages using the device escape code.
 */
	typedef struct {
		Byte device_code;
		Byte device_command;
		/* command-specific data follows */
	} NM_device_request_header;

	typedef struct {
		Byte code;	/* NM_device_esc */
		NM_device_request_header devReqHdr;
	} NM_device_request;

/*
 * Device Escape Product Query
 */
#define PRODUCT_QUERY_PRODUCT_CODE 1
#define PRODUCT_QUERY_COMMAND 1
	typedef struct {
		Byte code;	/* NM response code */
		Byte product;
		Byte model;	/* product specific values */
		Byte version;	/* product/model specific values */
		Byte product_specific;	/* product specific interpretation */
		Byte transceiver_ID;
//#ifdef BITF_LITTLE_ENDIAN
//    bits        transceiver_ID : 5;
//    bits        XID_reserved : 3;
//#else
//    bits        XID_reserved : 3;
//    bits        transceiver_ID : 5;
//#endif
	} NM_product_query_response;

	typedef struct {
#ifdef BITF_LITTLE_ENDIAN
		bits max_wait:3;
		bits nm_auth:1;
		bits rcv_timer:4;
#else
		bits rcv_timer:4;
		bits nm_auth:1;
		bits max_wait:3;
#endif
	} nmMsgParms;

	typedef enum nm_model_code {
		/*  0 */ NEURON_3150_CODE = 0,
		/*  1 */ NEURON_PL3150_CODE = 1,
		/*  2 */ NEURON_CY7C53150_CODE = 2,
		/*  8 */ NEURON_3120_CODE = 8,
		/*  9 */ NEURON_3120E1_CODE = 9,
		/* 10 */ NEURON_3120E2_CODE = 10,
		/* 11 */ NEURON_3120E3_CODE = 11,
		/* 12 */ NEURON_3120A20_CODE = 12,
		/* 13 */ NEURON_3120E5_CODE = 13,
		/* 14 */ NEURON_3120E4_CODE = 14,
		/* 15 */ NEURON_PL3120E4_CODE = 15,
		/* 16 */ NEURON_CY7C53120_CODE = 0x10,
		/* 32 */ NEURON_FT5000_CODE = 0x20,
		/* 33 */ NEURON_N5000_CODE = 0x21,
		/* 34 */ NEURON_PL5000_CODE = 0x22,
	} nm_model_code;

	typedef enum {
		MODEL_3150 = 0x00,
		MODEL_PL3150 = 0x01,
		MODEL_CY7C53150 = 0x02,	// 3150L

		MODEL_3120 = 0x08,
		MODEL_3120E1 = 0x09,
		MODEL_3120E2 = 0x0A,
		MODEL_3120E3 = 0x0B,
		MODEL_3120A20 = 0x0C,
		MODEL_3120E5 = 0x0D,
		MODEL_3120E4 = 0x0E,	// FT 3120
		MODEL_PL3120E4 = 0x0F,
		MODEL_CY7C53120 = 0x10,	// 3120L8

		MODEL_FT5000 = 0x20,
		MODEL_N5000 = 0x21,
		MODEL_PL5000 = 0x22,
		MODEL_ORION = 0x82
	} model_t;

#ifdef __cplusplus
}
#endif				/* __cplusplus */
#ifdef _MSC_VER
#pragma pack()
#endif
#if defined LINUX
#pragma pack()
#endif
#endif
