#ifndef _NI_MGMT2_H
#define _NI_MGMT2_H

#if defined LINUX

#include "ni_comm.h"
#include "ns_plat.h"

#pragma pack(1)

#endif

#ifdef _MSC_VER
#pragma pack(_NSS_PACKING)
#endif

/*
 *            NI_MGMT.H
 *
 * Network management codes and messages for an application node.
 * This file contains a subset of the network management structures
 * The types have been changed where necessary because of the difference
 * in representations between Neuron and PC-DOS objects.
 */

#define NM_update_nv_cnfg        0x6B
#define NM_update_nv_cnfg_fail   0x0B
#define NM_update_nv_cnfg_succ   0x2B
/*
typedef struct  {
    Byte            code;
    Byte            index;
    union {
        nv_struct       nv_cnfg;    // for index < 255
        struct {                    // for index >= 255
            Word    long_index;
            nv_struct   nv_cnfg;
        } escaped;
    } nvcnfg_union;
} NM_update_nv_cnfg_request;
*/
/*
typedef struct {
    Byte        code;
    nv_struct   nv_cnfg;
} NM_query_nv_cnfg_response;
*/
#define NM_query_nv_cnfg        0x68
#define NM_query_nv_cnfg_fail   0x08
#define NM_query_nv_cnfg_succ   0x28
/*
typedef struct {
    nv_struct   nv_cnfg;
    Byte        primary;
    Word        host_primary;
} alias_struct;
*/
typedef struct {
	Byte code;
	Byte index;
	alias_struct alias;
} NM_update_alias_request;

typedef struct {
	Byte code;
	alias_struct alias;
} NM_query_alias_response;

/*
typedef enum {
    APPL_OFFLINE = 0,
    APPL_ONLINE,
    APPL_RESET,
    CHANGE_STATE,
    DEFAULT_RESET
} nm_node_mode;
*/
/*
typedef enum {
    BLANK          = 0,
    INIT_CS_UNCFG  = 1,
    UNCONFIGURED   = 2,
    APPL_LESS      = 3,
    CONFIGURED     = 4,
    INIT_CS_CNFG   = 5,
    HARD_OFFLINE   = 6,     // Configured, hard offline
} nm_node_state;
*/
/*
typedef struct {
    Byte        code;
    Byte        mode;               // Interpret with 'nm_node_mode'
    Byte        node_state;         // Optional field if mode==CHANGE_STATE
                                    // Interpret with 'nm_node_state'
} NM_set_node_mode_request;
*/
#define NM_set_node_mode 0x6C
/*
typedef enum {
    ABSOLUTE_ADDR      = 0,
    READ_ONLY_RELATIVE = 1,
    CONFIG_RELATIVE    = 2,
    STATS_RELATIVE     = 3,
} nm_mem_mode;
*/
/*
typedef enum {
    NO_ACTION      = 0,
    BOTH_CS_RECALC = 1,
    CNFG_CS_RECALC = 4,
    NODE_RESET     = 8,
    FLASH_SYSTEM_IMAGE = 0x80
} nm_mem_form;
*/
/*
typedef struct {
    Byte    code;
    Byte    mode;
    Byte    offset_hi;
    Byte    offset_lo;
    Byte    count;
    Byte    form;                    // followed by the data
} NM_write_memory_request;
*/
#define NM_write_memory 0x6E
/*
typedef struct {
    Byte        code;
    Byte        mode;
    Byte        offset_hi;
    Byte        offset_lo;
    Byte        count;
} NM_read_memory_request;
*/
#define NM_read_memory 0x6D

// Network management message codes

#define NM_query_SNVT           0x72
#define NM_query_SNVT_fail      0x12
#define NM_query_SNVT_succ      0x32
/*
typedef struct {
    Byte    code;
    Byte    offset_hi;     // big-endian 16-bits
    Byte    offset_lo;
    Byte    count;
} NM_query_SNVT_request;
*/
#define NM_wink 0x70

#define NM_NV_fetch         0x73
#define NM_NV_fetch_fail    0x13
#define NM_NV_fetch_succ    0x33

typedef struct {
	Byte code;
	Byte nv_index;
	Word nv_index_16;
} NM_NV_index_hdr;

#define NV_INDEX_ESC 0xFF	/* NV index escape code */

#define DOMAIN_ID_LEN 6
#define AUTH_KEY_LEN  6
/*
typedef struct {
    Byte        id[ DOMAIN_ID_LEN ];
    Byte        subnet;
    bits        node        : 7;
    bits        cloneDomain : 1;    // this bit must be set to 1
    Byte        len;
    Byte        key[ AUTH_KEY_LEN ];
} domain_struct;
*/
/*
typedef struct  {
    Byte            code;
    Byte            domain_index;  // index;
    domain_struct   domain;
} NM_update_domain_request;
*/
#define NM_update_domain    0x63
/*
typedef struct  {
    Byte            code;
    domain_struct   domain;
} NM_query_domain_response;
*/
#define NM_query_domain     0x6a

typedef struct {
	Byte code;
	Word xmit_errors;	// stats are hi-Byte first
	Word transaction_timeouts;
	Word rcv_transaction_full;
	Word lost_msgs;
	Word missed_msgs;
	Byte reset_cause;
	bits state:3;		// interpret with nm_node_state
	bits offline:1;
	 bits:3;
	bits bypass:1;
	Byte version_number;
	Byte error_log;
	Byte model_number;
} ND_query_status_response;

typedef enum nm_proxy_sub {
	/* 0 */ QUERY_ID,
	/* 1 */ STATUS_REQUEST,
	/* 2 */ XCVR_STATUS
} nm_proxy_sub;

typedef struct ND_proxy_request {
	Byte sub_command;	// nm_proxy_sub
	Byte type;		// addr_type or (0x80 | group_size)
#ifdef BITF_LITTLE_ENDIAN
	bits member_or_node:7;
	bits dummy:1;
	bits retry:4;
	bits rpt_timer:4;
	bits tx_timer:4;
	bits rcv_timer:4;
#else
	bits dummy:1;
	bits member_or_node:7;
	bits rpt_timer:4;
	bits retry:4;
	bits rcv_timer:4;
	bits tx_timer:4;
#endif
	Byte group_or_subnet;
	Byte neuron_id[NEURON_ID_LEN];	// for type == 2
} ND_proxy_request;

typedef struct ND_query_xcvr_response {
	Byte xcvr_params[NUM_COMM_PARAMS];
} ND_query_xcvr_response;

typedef struct extended_version_number {
	Word manufacturer_id_high;
#ifdef BITF_LITTLE_ENDIAN
	// PC
	bits version_high:4;
	bits manufacturer_id_low:4;
#else
	//Neuron
	bits manufacturer_id_low:4;
	bits version_high:4;
#endif
	Byte version_low;
} extended_version_number;

typedef struct extended_self_definition {
#ifdef BITF_LITTLE_ENDIAN
	bits embedded_xcvr:4;
	bits version:4;
	bits clk_rate_max:3;
	bits sector_size:5;
#else
	bits version:4;
	bits embedded_xcvr:4;
	bits sector_size:5;
	bits clk_rate_max:3;
#endif
	Byte ext_flash_start;
	Byte ext_flash_end;
#ifdef BITF_LITTLE_ENDIAN
	bits ee_page_cnt:4;
	bits reserved1:4;
#else
	bits reserved1:4;
	bits ee_page_cnt:4;
#endif
	Byte ram_page_cnt;
	Byte flash_blk_size;
	Byte reserved2[3];
} extended_self_definition;

typedef struct ND_query_status_extended_response {
	extended_version_number evn;
	Byte minor_model;
	extended_self_definition sdi;
} ND_query_status_extended_response;

typedef struct ND_query_firmware_response {
	Byte firmware_version;
	Byte build_number;
} ND_query_firmware_response;

#define ND_query_status             0x51
#define ND_proxy		            0x52

#define ND_clear_status             0x53
#define ND_clear_status_fail        0x13
#define ND_clear_status_succ        0x33

#define ND_query_xcvr	            0x54
#define ND_query_status_extended    0x55

#define ND_query_firmware           0x59

typedef struct {
	Byte type;		// addr_type or (0x80 | group_size)
	bits member_or_node:7;
	bits domain:1;
	bits retry:4;
	bits rpt_timer:4;
	bits tx_timer:4;
	bits rcv_timer:4;
	Byte group_or_subnet;
} addr_struct;

typedef struct {
	Byte code;
	addr_struct addr;
} NM_query_addr_response;

#define NM_query_addr           0x67
#define NM_query_addr_succ      0x27
#define NM_query_addr_fail      0x07

typedef struct {
	Byte code;
	Byte index;
	addr_struct addr;
} NM_update_addr_request;

#define NM_update_addr          0x66
#define NM_update_group_addr    0x69

#define NM_checksum_recalc      0x6F

typedef enum {
	BOTH_CS = 1,
	CNFG_CS = 4,
} nm_checksum_recalc;

typedef struct {
	Byte code;
	Byte neuron_id[NEURON_ID_LEN];
	Byte id_string[ID_STR_LEN];
} NM_service_pin_msg;

#define NM_service_pin  0x7F

typedef enum {
	UNCONFIGD = 0,
	SELECTED = 1,
	SELECTED_UNCFG = 2,
} nm_query_selector;

typedef struct {
	Byte code;
	Byte selector;		// interpret with nm_query_selector
	Byte mode;		// interpret with nm_mem_mode
	Word offset;
	Byte count;		// followed by data to match
} NM_query_id_request;

#define NM_query_id        0x61
#define NM_query_id_fail   0x01
#define NM_query_id_succ   0x21

typedef struct {
	Byte code;
	Byte enable;		// interpret with boolean
} NM_respond_to_query_request;

#define NM_respond_to_query 0x62

typedef NM_service_pin_msg NM_query_id_response;

#define NM_anm_escape   0x7D

#define NM_memory_refresh              0x71
#define NM_memory_refresh_fail         0x11
#define NM_memory_refresh_success      0x31

typedef enum {
	ON_chip = 0,
	OFF_chip = 1
} nm_memory_refresh_which;

typedef struct {
	Byte code;
	NetWord offset;
	Byte count;
	Byte which;
} NM_memory_refresh_request;

// Define application network management code

typedef enum ANM_slta_code {
	ANM_product_query = 1,
} ANM_slta_code;

typedef struct ANM_slta_gen_request {
	Byte code;
	Byte sub_code;		// use enum ANM_sub_code
	Byte app_command;	// use enum ANM_slta_code
} ANM_slta_gen_request;

typedef enum {
	subc_slta = 1,
	subc_nss_xx = 2,
	subc_nss_yy = 3,
	subc_larva = 4,
} ANM_product;

typedef struct ANM_product_query_response {
	Byte code;
	Byte product;		// Use ANM_product
	Byte model;		// Use xxx_model
	Byte version;
	Byte config;		// Use slta_config for SLTA
	Byte transceiver_ID;
//    bits    transceiver_ID      : 5;
//    bits    XID_reserved        : 3;
} ANM_product_query_response;

#define NM_rtr_escape   0x7E

#define ND_rtr_status		0x7C
#define ND_rtr_status_fail  0x1C
#define ND_rtr_status_succ  0x3C

typedef enum { NON_IP_TO_IP_ROUTER = 1, IP_NODE, IP_PROXY, IP_TO_IP_ROUTER } nd_node_type;	// 852 Node Type
typedef enum { CONFIGD_RTR, LEARNING, BRIDGE, REPEATER } nd_rtr_algorithm;
typedef enum { NORMAL, INIT_RTR_TABLE, TEMP_BRIDGE } nd_rtr_mode;

typedef struct {
	Byte code;
	Byte algorithm;
	Byte mode;
} ND_rtr_status_response;

typedef struct {
	Byte code;
	Byte mode;
} NM_rtr_mode_request;

#define NM_rtr_mode			0x74
#define NM_rtr_mode_fail	0x14
#define NM_rtr_mode_succ	0x34

typedef struct {
	Byte code;
	bits index_times_8:2;
	bits unused:4;
	bits domain_index:1;
	bits group_or_subnet:1;
} NM_rtr_table_clear_request;

#define NM_rtr_table_clear		0x75
#define NM_rtr_table_clear_fail	0x15
#define NM_rtr_table_clear_succ 0x35

typedef struct {
	Byte code;
	bits index_times_8:2;
	bits unused:4;
	bits domain_index:1;
	bits group_or_subnet:1;
	Byte table[8];
} NM_rtr_table_downld_request;

#define NM_rtr_table_downld			0x76
#define NM_rtr_table_downld_fail	0x16
#define NM_rtr_table_downld_succ	0x36

typedef struct {
	Byte code;
	bits index_times_8:2;
	bits unused:3;
	bits ram_or_eeprom:1;
	bits domain_index:1;
	bits group_or_subnet:1;
} NM_rtr_table_report_request;

#define NM_rtr_table_report			0x7B
#define NM_rtr_table_report_fail	0x1B
#define NM_rtr_table_report_succ	0x3B

typedef struct {
	Byte code;
	Byte table[8];
} NM_rtr_table_report_response;

#define NM_rtr_group_fwd		0x77
#define NM_rtr_group_fwd_fail	0x17
#define NM_rtr_group_fwd_succ	0x37

typedef struct {
	Byte code;
	bits ram_or_eeprom:1;
	bits unused2:5;
	bits domain_index:1;
	bits unused1:1;
	Byte group;
} NM_rtr_group_fwd_request;

#define NM_rtr_subnet_fwd		0x78
#define NM_rtr_subnet_fwd_fail	0x18
#define NM_rtr_subnet_fwd_succ	0x38

typedef struct {
	Byte code;
	bits ram_or_eeprom:1;
	bits unused2:5;
	bits domain_index:1;
	bits unused1:1;
	Byte subnet;
} NM_rtr_subnet_fwd_request;

#define NM_rtr_group_nofwd		0x79
#define NM_rtr_group_nofwd_fail	0x19
#define NM_rtr_group_nofwd_succ	0x39

typedef struct {
	Byte code;
	bits ram_or_eeprom:1;
	bits unused2:5;
	bits domain_index:1;
	bits unused1:1;
	Byte group;
} NM_rtr_group_nofwd_request;

#define NM_rtr_subnet_nofwd			0x7A
#define NM_rtr_subnet_nofwd_fail	0x1A
#define NM_rtr_subnet_nofwd_succ	0x3A

typedef struct {
	Byte code;
	bits ram_or_eeprom:1;
	bits unused2:5;
	bits domain_index:1;
	bits unused1:1;
	Byte subnet;
} NM_rtr_subnet_nofwd_request;

#define NA_appl_msg         0x00
#define NA_appl_offline     0x3F
#define NA_foreign_msg      0x40
#define NA_foreign_offline  0x4F

typedef struct {
	NetWord xmit_errors;	// number of CRC errors during packet reception
	NetWord transaction_timeouts;
	NetWord rcv_transaction_full;
	NetWord lost_msgs;
	NetWord missed_msgs;
} base_stats;

typedef struct {
	Word node_received;
	Word node_addressed;
	Word transmitted;
	Word retries;
	Word backlog_ovflo;
	Word late_acks;
	Word collisions;
	Byte eeprom_lock;
} network_stats;

typedef struct {
	base_stats base;
	network_stats network;
} StatusStruct;

#define INCR_STATS(x) {NET_SWAP_WORD(x);((x < (Word)0xFFFF)?x++:x);NET_SWAP_WORD(x);}

#define NM_update_key		0x65
#define NM_update_key_fail	0x05
#define NM_update_key_succ	0x25

typedef struct {
	Byte code;
	Byte domain_index;
	Byte key[AUTH_KEY_LEN];
} NM_update_key_request;

    // Function prototypes
/*
#ifndef NI_MSG_H
typedef struct {
    Byte filler[ NEURON_ID_LEN + 4 ];
} SendAddrDtl;
#endif
*/
/*
boolean query_statistics( const SendAddrDtl *,
    boolean is_3120, Byte FW_version, network_stats * );

extern boolean get_model( const SendAddrDtl *, Byte * model );

boolean get_fw_model( const SendAddrDtl *, Byte *fw_version, Byte *model );

extern boolean get_base_version( const SendAddrDtl *, Byte * version );

extern boolean update_domain_entry( const SendAddrDtl *, Byte index,
    const domain_struct * domain_entry );

extern boolean update_addr_entry( const SendAddrDtl * node_addr, Byte index,
    const addr_struct * addr_entry );

extern boolean update_nv_entry( const SendAddrDtl * node_addr, Word index,
    const nv_struct * nv_entry );

extern boolean update_alias_entry( const SendAddrDtl * node_addr, Word index,
    const alias_struct * alias );

extern boolean set_node_mode( const SendAddrDtl *, nm_node_mode );

extern boolean set_node_state( const SendAddrDtl *, nm_node_state, ServiceType );

extern boolean get_IDs( const SendAddrDtl *, Byte * neuron_id,
    Byte * appl_id );        // read neuron and application IDs

extern boolean read_memory(
    const SendAddrDtl   * node_addr,
    nm_mem_mode     mode,
    Word            offset,
    Byte            length,
    void          * buffer );

extern boolean write_memory(                    // without verification
    const SendAddrDtl   * node_addr,
    nm_mem_mode     mode,
    Word            offset,
    Byte            length,
    const void    * buffer,
    nm_mem_form     form );

extern boolean write_memory_read(               // with verification
    const SendAddrDtl   * node_addr,
    nm_mem_mode     mode,
    Word            offset,
    Byte            length,
    const void    * buffer,
    nm_mem_form     form );

extern boolean write_config_byte(               // with verification
    const SendAddrDtl * node_addr,
    Word            offset,
    Byte            value );

extern boolean update_xsum(
    const SendAddrDtl * node_addr, nm_checksum_recalc which );

extern boolean clear_status( const SendAddrDtl * node_addr );

extern boolean query_status( const SendAddrDtl * node_addr,
    ND_query_status_response   *  status_response );

extern boolean query_nv_table( const SendAddrDtl * node_addr,
    Word nv_index, nv_struct * nv_cnfg );

extern boolean query_alias_table( const SendAddrDtl * node_addr,
    Word nv_index, alias_struct * alias );

extern boolean query_addr_table( const SendAddrDtl * node_addr,
    Byte addr_index, addr_struct * addr );

extern boolean query_domain_table( const SendAddrDtl * node_addr,
    Byte domain_index, domain_struct * domain );

extern boolean fetch_nv_data( const SendAddrDtl * node_addr, Word index,
    void * data, Byte * in_length );

extern boolean update_nv_data( const SendAddrDtl * node_addr,
    Byte selector_hi, Byte selector_lo, void * data, Byte length,
    boolean out_auth );

extern boolean get_nv_size( const SendAddrDtl * node_addr, Word index,
    Word nv_fixed_addr, Byte * nv_size );

extern boolean make_big_input_bufs( const SendAddrDtl * );
extern boolean make_big_output_bufs( const SendAddrDtl * );

extern boolean join_group( const SendAddrDtl *, Word group_number,
    Word group_size, Word member );
extern boolean set_non_group_timer( const SendAddrDtl * node_addr );
extern boolean select_nodes( const SendAddrDtl *, boolean select );
extern boolean query_id( const SendAddrDtl * node_addr, nm_query_selector,
    NM_query_id_response * response );
extern boolean find_nodes_in_domain( nm_query_selector selection );

extern boolean NI_configure( void );   // configure the network interface

//extern boolean write_buf_config( const SendAddrDtl *, const buf_config * );
//extern boolean read_buf_config( const SendAddrDtl *, buf_config * );
extern boolean get_app_config( void );

extern boolean read_host_SNVT(
    const SendAddrDtl   * node_addr,
    nm_mem_mode     mode,     // for compatibility with read_memory
    Word            offset,
    Byte            length,
    void          * buffer );

//void add_node( const NM_service_pin_msg * servpin_msg, node_type type );
*/
//extern SendAddrDtl neuron_addr;
//extern domain_struct NI_domain;

#if defined LINUX || defined _MSC_VER
#pragma pack()
#endif

#endif
