#ifndef _NS_PLAT_H
#define _NS_PLAT_H
/***************************************************************
 *  Filename: ns_plat.h
 *
 *  Description:  This file contains platform dependent NSS Service
 *                definitions.
 *
 *  Copyright (c) 1994-1997 Echelon Corporation.  All Rights Reserved.
 *
 ****************************************************************/

#ifdef _MSC_VER
#pragma pack(_NSS_PACKING)
#endif

#if defined LINUX
#pragma pack(1)
#endif

#ifdef __cplusplus
extern "C" {			/* Assume C declarations for C++ */
#endif				/* __cplusplus */

#define NULLVAL             0	/* non-pointer NULL value */
#define NS_NAME_LEN         17	/* general name length, including terminator */
#define NS_APPL_NAME_LEN    30	/* Length of NSS application name, including terminator */
#define NULL_NVMT_INDEX     0xFFFF	/* no NV or MT index specified */
#define MSG_IN_TAG          0xFFFE	/* indicates the msg_in tag */
#define MSG_OUT_TAG         0xFFFD	/* indicates a generic msg_out tag */
#define NVMT_INDEX_IS_MT    0x8000	/* bit indicates index is a msg tag */
#define NVMT_INDEX_IS_ALIAS 0x2000	/* indicates index is an NV alias */
#define NVMT_INDEX_ONLY_MASK 0x1FFF	/* mask off MT and alias flag bits */
#define NULL_LMOBJ_INDEX    0xFFFF	/* no LonMark object index specified */
#define NULL_LMOBJ_TYPE     0xFFFF	/* no LonMark object type specified */
#define CP_COUNT_UNKNOWN    0xFFFF	/* Indicates number of CPs is unknown */
#define NO_NV_IN_USE        (0xFFFF)	/* Value of maxNvInUse when no NVs defined */
#define NULL_MANAGER_HANDLE 0	/* Value to indicate no or unknown manager */
#define CHANNEL_HANDLE_UNKNOWN ((ChannelHandle)0)
	/* Value indicates that the channel handle
	 * is unknown.  NSS will determine the
	 * actual channel handle unless OFFNET
	 * or the target is on a repeater or
	 * permanent bridge-connected channel. */
#define CHANNEL_HANDLE_ARBITRARY ((ChannelHandle)-1)
	/* Value indicates that an arbitrary
	 * channel handle can be or was chosen.
	 * This is similar to
	 * CHANNEL_HANDLE_UNKNOWN except if the
	 * target is on a repeater or permanent
	 * bridge-connected channel, NSS will
	 * arbitrarily pick a channel from the
	 * set of possible candidates. Services
	 * which do channel selection will
	 * return this same value if an
	 * arbitrary selection was made.  In this
	 * case, the chosen channel can be
	 * retrieved using NS_GET_PROPERTY */

/* **********
   Typedefs derived from base types
   ********** */
	typedef Byte AddrTblIndex;	/* 0-14, address table index */
	typedef Byte AuthKey[AUTH_KEY_LEN];	/* authentication key */
	typedef NetWord ChannelHandle;	/* channel handle */
	typedef NetWord ChannelId;	/* channel id */
	typedef NetWord CpDimension;	/* configuration property (CP) dimension */
	typedef NetWord CpIndex;	/* CP index */
	typedef NetWord CpLength;	/* CP length */
	typedef NetWord CpField;	/* CP structure field */
	typedef NetWord DomainHandle;	/* domain handle */
	typedef Byte DomainId[DOMAIN_ID_LEN];	/* domain id (see also DomainSig) */
	typedef NetWord FileHandle;	/* file handle for file transfers */
	typedef NetWord FileIndex;	/* file index for file transfers */
	typedef NetWord LmObjIndex;	/* LonMark object index */
	typedef NetWord LmObjMemberIndex;	/* Members within LonMark obj (0..N-1) */
	typedef NetWord LmObjMemberNumber;	/* Member number in LonMark obj (as per guidelines) */
	typedef char LmObjName[LMOBJ_FULL_NAME_LEN];	/* LonMark object programmatic name */
	typedef NetWord LmObjType;	/* LonMark object type */
	typedef Byte Location[LOCATION_LEN];	/* location field in neuron */
	typedef NetWord ManagerHandle;	/* manager handle */
	typedef Byte NeuronId[NEURON_ID_LEN];	/* 48-bit unique neuron id */
	typedef Byte NodeId;	/* node number within subnet */
	typedef NetWord NvIndex;	/* 0-4095, usable as any nv index */
	typedef Byte MtIndex;	/* 0-14, any Mtag index */
	typedef NetWord NvMtIndex;	/* any NV or Mtag index */
	typedef NetWord NodeHandle;	/* global node handle */
	typedef char NsName[NS_NAME_LEN];	/* NSA name */
	typedef char NsApplName[NS_APPL_NAME_LEN];	/* NSS Application Name */
	typedef char NvName[NV_FULL_NAME_LEN];
	typedef NetWord NvSelector;	/* 14-bit nv selector */
	typedef NetWord ObjectCategory;	/* object category (class) */
	typedef Byte ProgramId[ID_STR_LEN];	/* program id */
	typedef NetWord Property;	/* object class */
	typedef NetWord RouterHandle;	/* router handle */
	typedef Byte SnvtType;	/* SNVT type */
	typedef NetWord SrSts;	/* error code */
	typedef NetWord SubnetHandle;	/* subnet handle */
	typedef Byte SubnetId;	/* subnet id */
	typedef NetWord SystemId;	/* system id */
	typedef NetWord ServiceCode;	/* service code */
	typedef NetWord ServiceLength;	/* Length, in bytes, of a service request */
	typedef NetWord TransactionHandle;	/* transaction handle */
	typedef NetWord TransactionId;	/* transaction id */
	typedef NetWord TransceiverId;	/* transceiver id */
	typedef Byte XcvrConfig[NUM_COMM_PARAMS];	/* transceiver configuration bytes */
	typedef Byte XcvrStatus[NUM_COMM_PARAMS];	/* neuron's transceiver parms */
	typedef ManagerHandle NssHandle;
	typedef ManagerHandle NsiHandle;
	typedef ManagerHandle SystemHandle;
	typedef ManagerHandle ApplHandle;
	typedef NetWord Version;	/* Object/property version number */

/*
** Definitions for object counts.  This assumes that there can not be
** more objects in a system than there are handles.
*/
	typedef ChannelHandle ChannelCount;
	typedef NetWord DeviceCount;	/* Device limit for channel or subnet */
	typedef DomainHandle DomainCount;
	typedef NodeHandle NodeCount;
	typedef RouterHandle RouterCount;
	typedef NetWord ProgramCount;
	typedef NvIndex NvCount;
	typedef Byte MtCount;
	typedef ManagerHandle ManagerCount;
	typedef NvIndex AliasCount;
	typedef LmObjIndex LmObjCount;
	typedef LmObjMemberIndex LmObjMemberCount;
	typedef CpIndex CpCount;

#ifdef ENHANCED_DOMAIN_FIELDS
#define DOMAIN_LEN_INVALID 0x01
#else
#define DOMAIN_LEN_INVALID 0x80
#define DOMAIN_LEN_MASK 0x07	// 3 Bits for dom_len
#define DOMAIN_OMA_MASK 0x18	// 2 Bits for dom_oma
#define DOMAIN_OMA_BIT  0x08	// 1 Bits for dom_oma = 1

	/* Typedef for the domain_struct returned by the function access_domain: */
#define NULL_DOMAIN_LEN 0xff
#endif

/* structure of a neuron domain entry.  Portable version of domain_struct */
	typedef struct DomainStruct {
		DomainId id;
		Byte subnet;
#ifdef BITF_LITTLE_ENDIAN
		BitField node:7;
		BitField cloneDomain:1;	// 0 => clone domain -> this bit must be set to 1
#else
		BitField cloneDomain:1;	// 0 => clone domain -> this bit must be set to 1
// The field "not_clone" is read-only, since it is managed within the
// update_domain() and update_clone_domain() functions, so there is no need
// to set it within the application program itself.  The value may be examined
// for the domain structure returned from the access_domain() function.
		BitField node:7;
#endif
#ifdef ENHANCED_DOMAIN_FIELDS
// Use of ENHANCED_DOMAIN_FIELDS and dom_oma is unsupported.
// If the symbol ENHANCED_DOMAIN_FIELDS is not defined (via #define),
// a backward-compatible definition of the domain_struct is provided.
#ifdef BITF_LITTLE_ENDIAN
		BitField dom_len:3;	// May only have values 0-6
		BitField dom_oma:2;	// 1 => open media auth (96-bit key)
		// honored even if domain addr invalid
		BitField dom_mbz:2;	// must be zero
		BitField dom_invalid:1;	// 1 => domain address is invalid
#else
		BitField dom_invalid:1;	// 1 => domain address is invalid
		BitField dom_mbz:2;	// must be zero
		BitField dom_oma:2;	// 1 => open media auth (96-bit key)
		// honored even if domain addr invalid
		BitField dom_len:3;	// May only have values 0-6
#endif
#else
		Byte len;	// May only have values 0-6
#endif
		AuthKey key;
	} DomainStruct, domain_struct;

	typedef struct DomainSig {	/* domain signature */
		DomainId id;
		Byte len;
	} DomainSig;

	typedef struct NodeSig {	/* node signature -- ordered */
		NeuronId neuronId;	/* exactly as the LONTalk */
		ProgramId programId;	/* service pin message */
	} NodeSig;

	typedef struct NodeAddr {	/* full node address */
		DomainId domainId;
		Byte len;
		Byte subnet;
		Byte node;
	} NodeAddr;

	typedef struct NvMtCounts {	/* Old form of node counts */
		NvCount nvCount;
		MtCount mtCount;
		NvCount maxNvInUse;
		NvCount maxNvSupported;
		NvCount staticNvCount;
	} NvMtCounts;

	typedef struct NodeCounts {	/* Objects associated with node */
		NvCount nvCount;
		MtCount mtCount;
		AliasCount aliasCount;
		LmObjCount lmObjCount;
		Bool cpAvailable;
		CpCount cpCount;
		NvCount maxNvInUse;
		NvCount maxNvSupported;
		NvCount staticNvCount;
	} NodeCounts;

	typedef struct NvCounts {	/* Objects associated with NV */
		CpCount cpCount;
	} NvCounts;

#ifdef BITF_LITTLE_ENDIAN
	typedef struct SubnetNodeAddr {	/* node address w/o domain id */
		Byte subnet;
		BitField node:7;
		 BitField:1;
	} SubnetNodeAddr;

#else
	typedef struct SubnetNodeAddr {	/* node address w/o domain id */
		Byte subnet;
		 BitField:1;
		BitField node:7;
	} SubnetNodeAddr;
#endif

	/* echelon internal note:  adjustments to
	   the value of NCLS_MAX must be
	   reflected in mchrdm.ddl */
	typedef enum _NodeClass {
		NCLS_DEBOUNCE,	/* node is being debounce; (not a class) */
		NCLS_MOBILE,	/* check detachment frequently (30 seconds) */
		NCLS_TEMPORARY,	/* check detachment less frequently (60 seconds) */
		NCLS_STATIONARY,	/* (default class) check detachment occassionally (5 minutes) */
		NCLS_PERMANENT,	/* never check */
		NCLS_MAX	/* max */
	} _NodeClass;
	typedef NetEnum(_NodeClass) NodeClass;

#define VALID_NODE_CLASS(nodeClass) ((nodeClass)>=NCLS_MOBILE && (nodeClass)<=NCLS_PERMANENT)

/* structures needed by declarations of property values: */

	typedef struct NvType {
		SnvtType snvt_type_index;	/* 0 means non-snvt             */
#ifdef BITF_LITTLE_ENDIAN
		BitField nv_service:2;	/* ACKD, UNACKD_RPT, UNACKD, REQUEST */
		 BitField:1;	/* not used                     */
		BitField nv_type_modifiable:1;	/* 1 = can change SNVT type     */
		BitField nv_dynamic:1;	/* 1 = Nv was dynamically added */
		BitField nv_auth:1;	/* 1 = uses authentication      */
		BitField nv_direction:1;	/* 1 = output                   */
		BitField nv_priority:1;	/* 1 = uses priority            */

		BitField nv_config_class:1;	/* 1 = config-class nv          */
		BitField nv_auth_config:1;	/* 1 = changeable via binding   */
		BitField nv_priority_config:1;	/* 1 = changeable via binding   */
		BitField nv_service_type_config:1;	/* 1 = changeable via binding   */
		BitField nv_offline:1;	/* 1 = update only while offline */
		BitField nv_polled:1;	/* 1 = polled nv                */
		BitField nv_sync:1;	/* 1 = synchronous              */
		BitField nv_has_desc_attribs:1;	/* 1 = this byte is valid       */
#else
		BitField nv_priority:1;	/* 1 = uses priority            */
		BitField nv_direction:1;	/* 1 = output                   */
		BitField nv_auth:1;	/* 1 = uses authentication      */
		BitField nv_dynamic:1;	/* 1 = Nv was dynamically added */
		BitField nv_type_modifiable:1;	/* 1 = can change SNVT type     */
		 BitField:1;	/* not used                     */
		BitField nv_service:2;	/* ACKD, UNACKD_RPT, UNACKD, REQUEST */

		BitField nv_has_desc_attribs:1;	/* 1 = this byte is valid       */
		BitField nv_sync:1;	/* 1 = synchronous              */
		BitField nv_polled:1;	/* 1 = polled nv                */
		BitField nv_offline:1;	/* 1 = update only while offline */
		BitField nv_service_type_config:1;	/* 1 = changeable via binding   */
		BitField nv_priority_config:1;	/* 1 = changeable via binding   */
		BitField nv_auth_config:1;	/* 1 = changeable via binding   */
		BitField nv_config_class:1;	/* 1 = config-class nv          */
#endif
		Byte nv_length;	/* length                       */

		/* Encoded message rates.  Zero means not available.        */
		/* Valid range is 1-127.  Actual value is:                  */
		/* (messages per seconds) = 2**((value)/8-5)                */
		Byte nv_rate_est;	/* estimated sustained rate     */
		Byte nv_max_rate_est;	/* estimated maximum rat        */
	} NvType;

/****************************************************************************
 * The following literals define the allocation of service, category, and prop
 * codes.
 ***************************************************************************/
#ifdef COMPILE_8		/* 8-bit compiler such as neuron c */
/* attributes: */
#define SVC_ATTRIB      (0xc000UL)	/* a mask to obtain the attribute value */
#define SVC_SERVICE     (0x0000UL)	/* a non-modifying service */
#define SVC_MODIFY      (0x4000UL)	/* a modifying service (requires NSS) */
#define SVC_EVENT       (0x8000UL)	/* an event */
#define SVC_RSV_ATTRIB  (0xc000UL)	/* reserved for future use */

/* scope: */
#define SVC_SCOPE       (0x3000UL)	/* a mask to obtain the scope value */
#define SVC_NSS         (0x0000UL)	/* an NSS API service */
#define SVC_LMK         (0x1000UL)	/* a LonMark standard service */
#define SVC_APP         (0x2000UL)	/* a proprietary service local to provider */
#define SVC_RSV_SCOPE   (0x3000UL)	/* reserved for future use */

/* additional scope of property: */
#define SVC_PGLOBAL     (0x0000UL)	/* property is trans-category */
#define SVC_PCAT        (0x4000UL)	/* property is category-specific */

/* additional scope of SrSts */
#define SRSTS_LCA       (0x4000L)	/* LCA SrSts */

#else				/* 16 and 32 bit compiler */
/* attributes: */
#define SVC_ATTRIB      (0xc000U)	/* a mask to obtain the attribute value */
#define SVC_SERVICE     (0x0000U)	/* a non-modifying service */
#define SVC_MODIFY      (0x4000U)	/* a modifying service (requires NSS) */
#define SVC_EVENT       (0x8000U)	/* an event */
#define SVC_RSV_ATTRIB  (0xc000U)	/* reserved for future use */

/* scope: */
#define SVC_SCOPE       (0x3000U)	/* a mask to obtain the class value */
#define SVC_NSS         (0x0000U)	/* an NSS API service */
#define SVC_LMK         (0x1000U)	/* a LonMark standard service */
#define SVC_APP         (0x2000U)	/* a proprietary service local to provider */
#define SVC_RSV_SCOPE   (0x3000U)	/* reserved for future use */

/* additional scope of property: */
#define SVC_PGLOBAL     (0x0000U)	/* property is trans-category */
#define SVC_PCAT        (0x4000U)	/* property is category-specific */

/* additional scope of SrSts */
#define SRSTS_LCA       (0x4000)	/* LCA SrSts */

#endif

/* these macros evaluate the attribute or scope and return TRUE if desired condition is met */
#define IS_SVC_NSS(c)       ((c & SVC_SCOPE)   == SVC_NSS)	/* ID is NSS-defined */
#define IS_SVC_LMK(c)       ((c & SVC_SCOPE)   == SVC_LMK)	/* ID is LonMark-defined */
#define IS_SVC_APP(c)       ((c & SVC_SCOPE)   == SVC_APP)	/* ID is application-defined */
#define IS_SVC_SERVICE(c)   ((c & SVC_ATTRIB)  == SVC_SERVICE)	/* Service code is a service */
#define IS_SVC_MODIFY(c)    ((c & SVC_ATTRIB)  == SVC_MODIFY)	/* Service code is modifying */
#define IS_SVC_EVENT(c)     ((c & SVC_ATTRIB)  == SVC_EVENT)	/* Service code is an event */
#define IS_SVC_PGLOBAL(p)   ((p & SVC_PCAT)    == SVC_PGLOBAL)	/* Property code is trans-category */
#define IS_SVC_PCAT(p)      ((p & SVC_PCAT)    == SVC_PCAT)	/* Property code is category-specific */

/* these macros evaluate the scope of an SrSts */
#define SRSTS_SCOPE     (0x7000UL)	/* a mask to obtain the scope */
#define IS_SRSTS_NSS(c)   (((c & SRSTS_SCOPE)   == SVC_NSS) \
                           && (c < 192 || c > 255))	/* ID is NSS-defined */
#define IS_SRSTS_LMK(c)   ((c & SRSTS_SCOPE)   == SVC_LMK)	/* ID is LonMark-defined */
#define IS_SRSTS_APP(c)   (((c & SRSTS_SCOPE)  == SVC_APP) \
                           || (c >= 192 && c <= 255))	/* ID is application-defined */
#define IS_SRSTS_LCA(c)   ((c & SRSTS_LCA)     == SRSTS_LCA)	/* ID is LCA-defined */

/* this macro returns the actual service number, without the attribute and scope */
#define SVC_CODE(fullCode) ((fullCode) & 0x0fff)
#define SERVICE_NUMBER  SVC_CODE
#define CATEGORY_NUMBER SVC_CODE
#define PROPERTY_NUMBER SVC_CODE
#define SRSTS_NUMBER    SVC_CODE

/* these symbols are used for backward compatibility: */
#define SVC_EVENT_FLAG  SVC_EVENT
#define SVC_APPL_FLAG   SVC_APP

/****************************************************************************
  Clients are represented by the typedef ClientId.  Client IDs are returned
  by the function NsCreateClient().
****************************************************************************/
	typedef Byte ClientId;

/* The #define CLIENT_ID_ALL is used to reference all clients. */
#define CLIENT_ID_ALL ((ClientId )-1)	/* 0xff */

/* The #define NO_CLIENT_ID is used to indicate that no client is defined for
 * a variable of type ClientId.
 */
#define NO_CLIENT_ID ((ClientId )-2)	/* 0xfe */

/****************************************************************************
  Servers are represented by the typedef ServerId.  Server IDs are
  allocated by the NSS API when a service request is received for the host
  and are freed by the API when the server calls NsSendResult().
****************************************************************************/
	typedef Byte ServerId;

/* The #define NO_SERVER_ID is used to indicate that no server is defined for
 * a variable of type ServerId.
 */
#define NO_SERVER_ID ((ServerId)NO_CLIENT_ID)	/* 0xfe */

/****************************************************************************
  This is a reserved value for a server handle to indicate that a service
  invocation is to be routed to my own NSS.
****************************************************************************/
#define MY_SYSTEM_MANAGER ((ManagerHandle) (0xffff))
#define MY_SYSTEM_MANAGER_ID MY_SYSTEM_MANAGER	/* obsolete -- for backward compatibility only */

/****************************************************************************
  DEFAULT_APPL_MANAGER is a reserved value of ApplHandle to indicate that the
  application is using the Application Manager Handle assigned to its NSI. When
  registering Service Request Handlers, if the current Appl Manager is set to
  DEFAULT_APPL_MANAGER (the default value), it will be treated as a wild card (i.e.,
  if there is not an exact match registered then it will match any specified
  application manager handle in the incoming service request.
****************************************************************************/
#define DEFAULT_APPL_MANAGER ((ApplHandle)(-1))	/* 0xffff */

/****************************************************************************
  STUB_APPL_MANAGER is a reserved value for an application that runs on top of

****************************************************************************/
#define STUB_APPL_MANAGER ((ApplHandle)(-2))	/* 0xfffe */

#ifdef __cplusplus
}
#endif				/* __cplusplus */
#ifdef _MSC_VER
#pragma pack()
#endif
#if defined LINUX
#pragma pack()
#endif
#endif
